/***************************************************************************
                         AC3_audio.c  -  description
                             -------------------
   Functions for handling AC3 sync frames.
                             -------------------
    begin                : Fri Aug 12 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*#define AC3_AUDIO_TRACE*/
#ifdef AC3_AUDIO_TRACE
# include <stdio.h>
#endif

#include "AC3_audio.h"
#include "bitstream.h"

static const u16 fsize480[] = {/* 48KHz values */
                                (64*2)-7,   (80*2)-7,   (96*2)-7,  (112*2)-7,
                               (128*2)-7,  (160*2)-7,  (192*2)-7,  (224*2)-7,
                               (256*2)-7,  (320*2)-7,  (384*2)-7,  (448*2)-7,
                               (512*2)-7,  (640*2)-7,  (768*2)-7,  (896*2)-7,
                              (1024*2)-7, (1152*2)-7, (1280*2)-7,},
                 fsize441[] = {/* 44.1KHz values */
                                (69*2)-7,   (87*2)-7,  (104*2)-7,  (121*2)-7,
                               (139*2)-7,  (174*2)-7,  (208*2)-7,  (243*2)-7,
                               (278*2)-7,  (348*2)-7,  (417*2)-7,  (487*2)-7,
                               (557*2)-7,  (696*2)-7,  (835*2)-7,  (975*2)-7,
                              (1114*2)-7, (1253*2)-7, (1393*2)-7,},
                 fsize320[] = {/* 32KHz values */
                                (96*2)-7,  (120*2)-7,  (144*2)-7,  (168*2)-7,
                               (192*2)-7,  (240*2)-7,  (288*2)-7,  (336*2)-7,
                               (384*2)-7,  (480*2)-7,  (576*2)-7,  (672*2)-7,
                               (768*2)-7,  (960*2)-7, (1152*2)-7, (1344*2)-7,
                              (1536*2)-7, (1728*2)-7, (1920*2)-7,};

#define DATA_PTR        ptr
#define END_DATA_PTR    end_pos
#define PARSER_STATE    frame->state


/* Find the first AC3 sync frame in the given data stream.
      INPUT: data_stream - Points to the data to look through.
             data_len    - Number of bytes in data_stream.
             frame       - Pointer to an AC3_SYNC_FRAME.
      RETURN: Success         - Returns number of bytes read from
                                  data_stream.
              Frame not found - 0 */
u32 AC3_AUDIO_next_frame(const u8       *data_stream,
                         u32             data_len,
                         AC3_SYNC_FRAME *frame)
{
   const u8 *ptr     = data_stream,
            *end_pos = data_stream + data_len;
   u8        data;

#undef  DATA_BYTE
#define DATA_BYTE       frame->data
#undef  BIT_SHIFT_BYTE
#define BIT_SHIFT_BYTE  frame->bit_shift
#undef  STOP_PARSING
#define STOP_PARSING    goto AC3_AUDIO_next_frame_END

#define FIND_SYNC_BYTE1    (AC3_AUDIO_next_frame_STATE+0)
#define FIND_SYNC_BYTE2    (AC3_AUDIO_next_frame_STATE+1)
#define SKIP_CRC           (AC3_AUDIO_next_frame_STATE+2)
#define CHECK_BSID         (AC3_AUDIO_next_frame_STATE+3)
#define GET_AC_MODE        (AC3_AUDIO_next_frame_STATE+4)
#define GET_LFEON          (AC3_AUDIO_next_frame_STATE+5)

   if (PARSER_STATE > AC3_AUDIO_next_frame_STATE) {
      switch (PARSER_STATE) {
      case FIND_SYNC_BYTE2:
         ptr--;
         goto AC3_AUDIO_next_frame_FIND_SYNC_BYTE2;
      case SKIP_CRC:
         ptr += frame->data;
         goto AC3_AUDIO_next_frame_SKIP_CRC;
      case CHECK_BSID:
         goto AC3_AUDIO_next_frame_CHECK_BSID;
      case GET_AC_MODE:
         goto AC3_AUDIO_next_frame_GET_AC_MODE;
      case GET_LFEON:
         goto AC3_AUDIO_next_frame_GET_LFEON;
      case AC3_AUDIO_SKIP_AUX_BYTES:
         ptr += frame->frame_len;
         if (ptr >= end_pos) {
            frame->frame_len = (u32)(ptr - end_pos);
            STOP_PARSING;
         }
      }
   }
   
AC3_AUDIO_next_frame_FIND_SYNC_BYTE1:
   /* look for sync word */
   do {
      while (*ptr != AC3_AUDIO_SYNC_BYTE1) {
         NEXT_BYTE(FIND_SYNC_BYTE1)
      }
AC3_AUDIO_next_frame_FIND_SYNC_BYTE2:
      NEXT_BYTE(FIND_SYNC_BYTE2)
   } while (*ptr != AC3_AUDIO_SYNC_BYTE2);

   /* skip CRC */
   ptr += 3;
AC3_AUDIO_next_frame_SKIP_CRC:
   if (ptr >= end_pos) {
      frame->data  = (u8)(ptr - end_pos);
      PARSER_STATE = SKIP_CRC;
      STOP_PARSING;
   }

   /* get sample rate, and frame length values */
   data = *ptr & 0x3F;
   if ( (*ptr >= 0xC0)||(data > 0x25) )
      goto AC3_AUDIO_next_frame_FIND_SYNC_BYTE1;   /* invalid sample rate */
   else if (*ptr >= 0x80) {
      frame->sample_rate = 32000;
      frame->frame_len   = (u32)fsize320[data >> 1];
   }
   else if (*ptr >= 0x40) {
      frame->sample_rate = 44100;
      frame->frame_len   = (u32)fsize441[data >> 1];
      if (data & 1)
         frame->frame_len += 2;
   }
   else {
      frame->sample_rate = 48000;
      frame->frame_len   = (u32)fsize480[data >> 1];
   }
#ifdef AC3_AUDIO_TRACE
   {
      u16 bitrate;
      
      data &= 0xFE;
      if (data <= 8)
         bitrate = ((u16)data << 2) + 32;
      else if (data <= 16)
         bitrate =  (u16)data << 3;
      else if (data <= 24)
         bitrate = ((u16)data << 4) - 128;
      else
         bitrate = ((u16)data << 5) - 512;
   printf("AC3_SYNC_FRAME:\n"
          "  bitrate=%huKbps  samplerate=%uHz  frame_len=%ubytes\n", 
          bitrate, frame->sample_rate, (frame->frame_len+7));
   }
#endif
   NEXT_BYTE(CHECK_BSID)

AC3_AUDIO_next_frame_CHECK_BSID:
   /* check bitstream ID, and get mode */
   if (*ptr > AC3_AUDIO_BSID)    /* ID is too high for this decoder */
      goto AC3_AUDIO_next_frame_FIND_SYNC_BYTE1;   
   frame->mode = *ptr & 0x07;
   NEXT_BYTE(GET_AC_MODE)

AC3_AUDIO_next_frame_GET_AC_MODE:
   /* get the audio mode */
   BIT_SHIFT_BYTE = 3;
   frame->mode   |= *ptr & 0xE0;
   if ((*ptr & 0xE0) == AC3_AUDIO_AC_20)
      BIT_SHIFT_BYTE += 2;       /* skip dsurmod */
   else {
      if ( (*ptr >= AC3_AUDIO_AC_30)&&(*ptr & 0x20) )
         BIT_SHIFT_BYTE += 2;    /* skip cmixlev */
      if (*ptr >= AC3_AUDIO_AC_21)
         BIT_SHIFT_BYTE += 2;    /* skip surmixlev */
   }

   if (*ptr >= AC3_AUDIO_AC_32)
      frame->num_ch = 5;
   else if (*ptr >= AC3_AUDIO_AC_31)
      frame->num_ch = 4;
   else if (*ptr >= AC3_AUDIO_AC_30)
      frame->num_ch = 3;
   else if (*ptr & AC3_AUDIO_AC_10)
      frame->num_ch = 1;
   else
      frame->num_ch = 2;
   ALIGN8_DATA_BYTE(GET_LFEON);

AC3_AUDIO_next_frame_GET_LFEON:
   /* get the lfeon bit */
   GET_LSB;
   if (DATA_BYTE >= 0x80)
      frame->mode |= AC3_AUDIO_AC_LFE;
#ifdef AC3_AUDIO_TRACE
   printf("  AC3_BSI:\n    bsmod=");
   switch (AC3_AUDIO_BS(frame)) {
   case AC3_AUDIO_BS_MS_CM:
      printf("Main Audio Service: Complete Main");
      break;
   case AC3_AUDIO_BS_MS_ME:
      printf("Main Audio Service: Music & Effects");
      break;
   case AC3_AUDIO_BS_AS_VI:
      printf("Associated Service: Visually Impaired");
      break;
   case AC3_AUDIO_BS_AS_HI:
      printf("Associated Service: Hearing Impaired");
      break;
   case AC3_AUDIO_BS_AS_D:
      printf("Associated Service: Dialogue");
      break;
   case AC3_AUDIO_BS_AS_C:
      printf("Associated Service: Commentary");
      break;
   case AC3_AUDIO_BS_AS_E:
      printf("Associated Service: Emergency");
      break;
   default:
      if (AC3_AUDIO_AC(frame) == AC3_AUDIO_AC_10)
         printf("Associated Service: Voice Over");
      else if (AC3_AUDIO_AC(frame) >= AC3_AUDIO_AC_20)
         printf("Main Audio Service: Karaoke");
      else
         printf("UNKNOWN");
   }
   printf("\n    acmod=");
   switch (AC3_AUDIO_AC(frame)) {
   case AC3_AUDIO_AC_1p1:
      printf("CH1 CH2 ");
      break;
   case AC3_AUDIO_AC_10:
      printf("CEN ");
      break;
   case AC3_AUDIO_AC_20:
      printf("LFT RGT");
      break;
   case AC3_AUDIO_AC_30:
      printf("LFT CEN RGT");
      break;
   case AC3_AUDIO_AC_21:
      printf("LFT RGT SUR");
      break;
   case AC3_AUDIO_AC_31:
      printf("LFT CEN RGT SUR");
      break;
   case AC3_AUDIO_AC_22:
      printf("LFT RGT LSUR RSUR");
      break;
   case AC3_AUDIO_AC_32:
      printf("LFT CEN RGT LSUR RSUR");
      break;
   default:
      printf("UNKNOWN");
   }
   if (frame->mode & AC3_AUDIO_AC_LFE)
      printf(" LFE");
   printf("  num_ch=%hu", (u16)frame->num_ch);
   if (frame->mode & AC3_AUDIO_AC_LFE)
      printf(".1");
   printf("\n");
#endif

   PARSER_STATE         = AC3_AUDIO_next_frame_END_STATE;
   frame->payload       = ptr;
   frame->payload_len16 = (u16)(data_len - (u32)(ptr - data_stream));
   return (u32)(ptr - data_stream);

AC3_AUDIO_next_frame_END:
   return 0;
#undef FIND_SYNC_BYTE1
#undef FIND_SYNC_BYTE2
#undef SKIP_CRC
#undef CHECK_BSID
#undef GET_AC_MODE
#undef GET_LFEON
}

/* Read the BSI data.
      INPUT: frame - Pointer to an AC3_SYNC_FRAME. */
void AC3_AUDIO_read_BSI(AC3_SYNC_FRAME *frame)
{
   const u8 *ptr       = frame->payload,
            *end_pos   = frame->payload + frame->payload_len16;
   u8        bit_shift = frame->bit_shift,
             data      = frame->data;

#undef  DATA_BYTE
#define DATA_BYTE       data
#undef  BIT_SHIFT_BYTE
#define BIT_SHIFT_BYTE  bit_shift
#undef  STOP_PARSING
#define STOP_PARSING    goto AC3_AUDIO_read_BSI_END

#define SKIP_COMPR      (AC3_AUDIO_read_BSI_STATE+ 1)
#define SKIP_LANGCOD    (AC3_AUDIO_read_BSI_STATE+ 2)
#define SKIP_AUDPRODI   (AC3_AUDIO_read_BSI_STATE+ 3)
#define SKIP_CH2DATA    (AC3_AUDIO_read_BSI_STATE+ 4)
#define SKIP_COMPR2     (AC3_AUDIO_read_BSI_STATE+ 5)
#define SKIP_LANGCOD2   (AC3_AUDIO_read_BSI_STATE+ 6)
#define SKIP_AUDPRODI2  (AC3_AUDIO_read_BSI_STATE+ 7)
#define SKIP_COPYRIGHT  (AC3_AUDIO_read_BSI_STATE+ 8)
#define SKIP_TIMECOD1   (AC3_AUDIO_read_BSI_STATE+ 9)
#define SKIP_TIMECOD1b  (AC3_AUDIO_read_BSI_STATE+10)
#define SKIP_TIMECOD2   (AC3_AUDIO_read_BSI_STATE+11)
#define SKIP_TIMECOD2b  (AC3_AUDIO_read_BSI_STATE+12)
#define SKIP_ADDBSI     (AC3_AUDIO_read_BSI_STATE+13)
#define SKIP_ADDBSIb    (AC3_AUDIO_read_BSI_STATE+14)
#define SKIP_ADDBSIc    (AC3_AUDIO_read_BSI_STATE+15)

   if (PARSER_STATE > AC3_AUDIO_read_BSI_STATE) {
      switch (PARSER_STATE) {
      case SKIP_COMPR:
         goto AC3_AUDIO_read_BSI_SKIP_COMPR;
      case SKIP_LANGCOD:
         goto AC3_AUDIO_read_BSI_SKIP_LANGCOD;
      case SKIP_AUDPRODI:
         goto AC3_AUDIO_read_BSI_SKIP_AUDPRODI;
      case SKIP_CH2DATA:
         goto AC3_AUDIO_read_BSI_SKIP_CH2DATA;
      case SKIP_COMPR2:
         goto AC3_AUDIO_read_BSI_SKIP_COMPR2;
      case SKIP_LANGCOD2:
         goto AC3_AUDIO_read_BSI_SKIP_LANGCOD2;
      case SKIP_AUDPRODI2:
         goto AC3_AUDIO_read_BSI_SKIP_AUDPRODI2;
      case SKIP_COPYRIGHT:
         goto AC3_AUDIO_read_BSI_SKIP_COPYRIGHT;
      case SKIP_TIMECOD1:
         goto AC3_AUDIO_read_BSI_SKIP_TIMECOD1;
      case SKIP_TIMECOD1b:
         goto AC3_AUDIO_read_BSI_SKIP_TIMECOD1b;
      case SKIP_TIMECOD2:
         goto AC3_AUDIO_read_BSI_SKIP_TIMECOD2;
      case SKIP_TIMECOD2b:
         goto AC3_AUDIO_read_BSI_SKIP_TIMECOD2b;
      case SKIP_ADDBSI:
         goto AC3_AUDIO_read_BSI_SKIP_ADDBSI;
      case SKIP_ADDBSIb:
         goto AC3_AUDIO_read_BSI_SKIP_ADDBSIb;
      case SKIP_ADDBSIc:
         goto AC3_AUDIO_read_BSI_SKIP_ADDBSIc;
      }
   }

   /* All data here is currently ignored! */
   if (DATA_BYTE & 0x02) {
      ALIGN8_DATA_BYTE(SKIP_COMPR);
AC3_AUDIO_read_BSI_SKIP_COMPR:
      GET_LSB;
   }
   if (DATA_BYTE & 0x01) {
      ALIGN8_DATA_BYTE(SKIP_LANGCOD);
   }
AC3_AUDIO_read_BSI_SKIP_LANGCOD:
   ALIGN8_DATA_BYTE(SKIP_AUDPRODI);
AC3_AUDIO_read_BSI_SKIP_AUDPRODI:
   GET_LSB;
   if (DATA_BYTE >= 0x80) {
      ALIGN_DATA_BYTE(SKIP_CH2DATA, 7);
AC3_AUDIO_read_BSI_SKIP_CH2DATA:
      GET_LSB;
   }

   if (AC3_AUDIO_AC(frame) == AC3_AUDIO_AC_1p1) {
      if (DATA_BYTE & 0x02) {
         ALIGN8_DATA_BYTE(SKIP_COMPR2);
AC3_AUDIO_read_BSI_SKIP_COMPR2:
         GET_LSB;
      }
      if (DATA_BYTE & 0x01) {
         ALIGN8_DATA_BYTE(SKIP_LANGCOD2);
      }
AC3_AUDIO_read_BSI_SKIP_LANGCOD2:
      ALIGN8_DATA_BYTE(SKIP_AUDPRODI2);
AC3_AUDIO_read_BSI_SKIP_AUDPRODI2:
      GET_LSB;
      if (DATA_BYTE >= 0x80) {
         ALIGN_DATA_BYTE(SKIP_COPYRIGHT, 7);
AC3_AUDIO_read_BSI_SKIP_COPYRIGHT:
         GET_LSB;
      }
   }

   if (DATA_BYTE & 0x10) {
      ALIGN8_DATA_BYTE(SKIP_TIMECOD1);
AC3_AUDIO_read_BSI_SKIP_TIMECOD1:
      ALIGN_DATA_BYTE(SKIP_TIMECOD1b, 6);
AC3_AUDIO_read_BSI_SKIP_TIMECOD1b:
      GET_LSB;   
   }

   if (DATA_BYTE & 0x08) {
      ALIGN8_DATA_BYTE(SKIP_TIMECOD2);
AC3_AUDIO_read_BSI_SKIP_TIMECOD2:
      ALIGN_DATA_BYTE(SKIP_TIMECOD2b, 6);
AC3_AUDIO_read_BSI_SKIP_TIMECOD2b:
      GET_LSB;
   }
   
   frame->tmp_data = (u32)(DATA_BYTE & 0x04);
   ALIGN_DATA_BYTE(SKIP_ADDBSI, 6);
AC3_AUDIO_read_BSI_SKIP_ADDBSI:
   if (frame->tmp_data) {
      frame->tmp_data = (u32)(DATA_BYTE >> 2) + 1;
      ALIGN_DATA_BYTE(SKIP_ADDBSIb, 6);
AC3_AUDIO_read_BSI_SKIP_ADDBSIb:
      do {
         ALIGN8_DATA_BYTE(SKIP_ADDBSIc);
AC3_AUDIO_read_BSI_SKIP_ADDBSIc:
         frame->tmp_data--;
      } while (frame->tmp_data);
   }
   GET_LSB;

   frame->bit_shift      = bit_shift;
   frame->data           = data;
   PARSER_STATE          = AC3_AUDIO_read_BSI_END_STATE;
   frame->payload_len16 -= (u16)(ptr - frame->payload);
   frame->payload        = ptr;
   return;

AC3_AUDIO_read_BSI_END:
   frame->bit_shift     = bit_shift;
   frame->data          = data;
   frame->payload_len16 = 0;
#undef SKIP_COMPR
#undef SKIP_LANGCOD
#undef SKIP_AUDPRODI
#undef SKIP_CH2DATA
#undef SKIP_COMPR2
#undef SKIP_LANGCOD2
#undef SKIP_AUDPRODI2
#undef SKIP_COPYRIGHT
#undef SKIP_TIMECOD1
#undef SKIP_TIMECOD1b
#undef SKIP_TIMECOD2
#undef SKIP_TIMECOD2b
#undef SKIP_ADDBSI
#undef SKIP_ADDBSIb
#undef SKIP_ADDBSIc
}

#undef DATA_BYTE
#undef BIT_SHIFT_BYTE
#undef DATA_PTR
#undef END_DATA_PTR
#undef PARSER_STATE
#undef STOP_PARSING

