/***************************************************************************
                          PMT_module.c  -  description
                             -------------------
   PMT PSI module functions.
                             -------------------
    begin                : Tue Aug 24 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/* #include <stdio.h> */

#include "MPEG2_SYS.h"
#include "PSI_SECTION.h"
#include "PMT_module.h"

/*int PMT_MODULE_PRINT_DATA;*/


/* PRIVATE STRUCTURES */
typedef struct {
   u8           desc_size,    /*     Number of valid pointers in the desc
                                       array. */
                elem_size;    /*     Nuber of valid pointers in the elem array. */
   u16          PCR_PID;      /* 13: The PID of the TS packets carrying
                                       the PCR for this program.
                        <0x1FFF> Used when there is no PCR. */
   PMT_DESC    *desc[PMT_MAX_PROG_DESC];
   ELEM_STREAM *elem[MPEG2_SYS_PIDS-2];
} FULL_PMT_SECTION;

typedef struct {
   u16              version, prog_num, active, changed;
   FULL_PMT_SECTION pmt;
   ELEM_STREAM      elem[MPEG2_SYS_PIDS-2];
   u8               prog_desc[PMT_MAX_PROG_DESC * sizeof(PMT_DESC)],
                    elem_desc[((MPEG2_SYS_PIDS-2)*PMT_MAX_ELEM_DESC) * sizeof(PMT_DESC)];
} PMT_module_data;

/* LOCAL FUNCTIONS */
static int PMT_SECTION_read(const u8        *data_stream,
                            int              stream_len,
                            PMT_module_data *pmt_data);
static int PMT_DESC_read(const u8  *data_stream,
                         int        stream_len,
                         PMT_DESC **desc_ptr,
                         u8        *desc,
                         int        max_desc);
static int ELEM_STREAM_read(const u8    *data_stream,
                            int          stream_len,
                            ELEM_STREAM *elem,
                            u8          *desc);
/*
static void PMT_display(PMT_SECTION *pmt, u16 ver, u16 prog_num);
*/
                                                                                                    
/*****************************************************************
   GENERAL MODULE FUNCTIONS
     These functions are used by all SYSTEM, PSI, and PES modules.
 *****************************************************************/
/* Returns the number of bytes of memory needed by the module. */
size_t PMT_PSI_MODULE_mem_needed()
{
   return sizeof(PMT_module_data);
}

/* Initializes the memory provided for the module.
      INPUT: void *     - Pointer to the memory provided for the module.
             SYS_DATA * - Pointer to system data. */
void PMT_PSI_MODULE_init(void *mem, SYS_DATA *sys_data)
{
   PMT_PSI_MODULE_done(mem);
}

/* Clean up (DO NOT FREE) the memory provided for the module.
      INPUT: void * - Pointer to the memory provided for the module. */
void PMT_PSI_MODULE_done(void *mem)
{
   int              i;
   PMT_module_data *data = (PMT_module_data *)mem;

   data->version = data->prog_num = data->active = 0;
   data->changed = 1;

   for (i = 0; i < (MPEG2_SYS_PIDS-2); i++) {
      data->pmt.elem[i] = data->elem + i;
   }
}


/**************************************************************************
   COMMON PSI MODULE FUNCTIONS
     These functions are used by SYSTEM modules to process the PSI_SECTIONs
       they produce.
 **************************************************************************/
/* Processes the packet handled by the module.
      INPUT: void * - Pointer to the memory provided for the module.
             void * - Pointer to the PSI_SECTION that needs handling. */
void PMT_PSI_MODULE_process_pkt(void *mem, void *psi)
{
   PMT_module_data *data    = (PMT_module_data *)mem;
   PSI_SECTION     *psi_sec = (PSI_SECTION *)psi;
   u16              psi_ver = (u16)PSI_SECTION_VERSION_NUMBER(psi_sec);

   if ( (psi_sec->table_id != PSI_SECTION_PROGRAM_MAP_ID)||
        (!PSI_SECTION_CURRENT_NEXT_FLAG(psi_sec)) )
      return;  /* PSI_SECTION does not contain PMT data, or PMT data is not current */
   else if (!data->active) {
      data->version  = psi_ver;
      data->prog_num = psi_sec->table_id_extension;
   }
   else if (data->prog_num != psi_sec->table_id_extension) {
      /* id extension value does not match, so clear table */
      data->version = data->prog_num = data->active = 0;
      data->changed = 1;
      return;
   }
   else if (data->version == psi_ver)
      return;  /* PMT data is already loaded */

   /* read PMT data */
   if (PMT_SECTION_read(psi_sec->payload,
                        (int)psi_sec->payload_len16,
                        data)) {
      /* activate PMT data */
      data->active = data->changed = 1;
/*
      if (PMT_MODULE_PRINT_DATA)
         PMT_display((PMT_SECTION *)&(data->pmt), data->version, data->prog_num);
*/
      return;
   }
   else if (data->active) {
      /* de-activate PMT data */
      data->active  = 0;
      data->changed = 1;
   }
}


/*************************************************************************
   PMT MODULE FUNCTIONS
     These functions are used by the MPEG2 TS decoder to get the PMT data.
       The PAT data is extracted from the PSI_SECTIONs sent by a system
       module.
 *************************************************************************/
/* Returns a non-zero value if the PMT data has changed, since the last
     time this function was called.
      INPUT: void * - Pointer to the memory provided for the module. */
int PMT_PSI_MODULE_PMT_changed(void *mem)
{
   PMT_module_data *data = (PMT_module_data *)mem;

   if (data->changed) {
      data->changed = 0;
      return 1;
   }
   return 0;
}

/* Adds an entry to the PMT, using the given values.  This function is
     used when there is no PMT data in the stream.
      INPUT: void * - Pointer to the memory provided for the module.
             u16    - Program number of the new entry. Same value used
                        in the MPEG2_PAT_MODULE_add_entry() function.
             u16    - PID of the PCR.
             u16    - PID of the video stream.
             u8     - One of the STREAM_TYPE_* values defined
                        in PMT_SECTION.h
             u16    - PID of the audio stream. 
             u8     - One of the STREAM_TYPE_* values defined
                        in PMT_SECTION.h */
void PMT_PSI_MODULE_add_entry(void *mem,
                              u16   program_number,
                              u16   PCR_PID,
                              u16   video_PID,
                              u8    video_type,
                              u16   audio_PID,
                              u8    audio_type)
{
   PMT_module_data *data = (PMT_module_data *)mem;

   data->version  = 0x00FF;
   data->prog_num = program_number;
   data->active   = 1;
   data->changed  = 1;

   data->pmt.PCR_PID   = PCR_PID;
   data->pmt.desc_size = 0;
   data->pmt.elem_size = 0;
   
   if (video_PID < TS_PKT_NUL_ID) {
      data->elem[data->pmt.elem_size].stream_type    = video_type;
      data->elem[data->pmt.elem_size].desc_size      = 0;
      data->elem[data->pmt.elem_size].elementary_PID = video_PID;
      
      data->pmt.elem[data->pmt.elem_size] = data->elem + data->pmt.elem_size;
      data->pmt.elem_size++;
   }

   if (audio_PID < TS_PKT_NUL_ID) {
      data->elem[data->pmt.elem_size].stream_type    = audio_type;
      data->elem[data->pmt.elem_size].desc_size      = 0;
      data->elem[data->pmt.elem_size].elementary_PID = audio_PID;

      data->pmt.elem[data->pmt.elem_size] = data->elem + data->pmt.elem_size;
      data->pmt.elem_size++;
   }
/*
   if (PMT_MODULE_PRINT_DATA)
      PMT_display((PMT_SECTION *)&(data->pmt), data->version, data->prog_num);
*/
}

/* Returns a pointer to the PMT_SECTION for the current program.
      INPUT: void * - Pointer to the memory provided for the module.
      RETURN: The whole PMT_SECTION is stored - Pointer to a PMT_SECTION.
              No PMT_SECTION is stored        - NULL */
PMT_SECTION *PMT_PSI_MODULE_get_PMT(void *mem)
{
   PMT_module_data *data = (PMT_module_data *)mem;

   if (!data->active)
      return NULL;
   return (PMT_SECTION *)&(data->pmt);
}


static int PMT_SECTION_read(const u8        *data_stream,
                            int              stream_len,
                            PMT_module_data *pmt_data)
{
   int rc;
   u16 desc_len;
   
   pmt_data->pmt.PCR_PID  = (u16)(*data_stream & 0x1F) << 8;
   data_stream++;

   pmt_data->pmt.PCR_PID |= (u16)(*data_stream);
   data_stream++;

   desc_len  = (u16)(*data_stream & 0x0F) << 8;
   data_stream++;

   desc_len |= (u16)(*data_stream);
   data_stream++;

   stream_len -= (int)desc_len + 4;
   if (stream_len >= 0) {
      /* read program descriptors */
      rc = PMT_DESC_read(data_stream, (int)desc_len, pmt_data->pmt.desc,
                         pmt_data->prog_desc, PMT_MAX_PROG_DESC);
      if (rc < 0)
         goto PMT_SECTION_read_error;
      else if (rc == 0)
         pmt_data->pmt.desc_size = 0;
      else {
         pmt_data->pmt.desc_size = (u8)rc;
         data_stream            += desc_len;
      }

      /* read program elements */
      rc = ELEM_STREAM_read(data_stream, stream_len, pmt_data->elem,
                            pmt_data->elem_desc);
      if (rc < 0)
         goto PMT_SECTION_read_error;

      pmt_data->pmt.elem_size = (u8)rc;
      return 1;   /* everything is ok */
   }

PMT_SECTION_read_error:      
   pmt_data->pmt.desc_size = pmt_data->pmt.elem_size = 0;
   return 0;   /* stream_len value is not right */
}

static int PMT_DESC_read(const u8  *data_stream,
                         int        stream_len,
                         PMT_DESC **desc_ptr,
                         u8        *desc,
                         int        max_desc)
{
   int num_desc = 0;

   while (stream_len >= 2) {
      *desc_ptr = (PMT_DESC *)desc;
      
      (*desc_ptr)->desc_tag = *data_stream;
      data_stream++;

      (*desc_ptr)->data[0]     = *data_stream;
      (*desc_ptr)->desc_length = *data_stream;
      data_stream             += (int)(*desc_ptr)->desc_length + 1;
      stream_len              -= (int)(*desc_ptr)->desc_length + 2;
      
/*      if ((*desc_ptr)->desc_length < 2) { */
         (*desc_ptr)->desc_length = 2;
         desc                    += 4;
/*      }
      else
         desc                 += (int)(*desc_ptr)->desc_length + 2; */
      desc_ptr++;
      num_desc++;

      if ( (num_desc == max_desc)&&(stream_len >= 2) ) {
         stream_len = 0;
/*
         if (PMT_MODULE_PRINT_DATA)
            printf("PMT_DESC_read: Need more than %d descriptors\n", max_desc);
*/
      }
   }
   return ((stream_len == 0) ? num_desc : -1);
}

static int ELEM_STREAM_read(const u8    *data_stream,
                            int          stream_len,
                            ELEM_STREAM *elem,
                            u8          *desc)
{
   int num_elem = 0,
       rc;
   u16 desc_len;
   
   while (stream_len >= 5) {
      elem->stream_type = *data_stream;
      data_stream++;

      elem->elementary_PID  = (u16)(*data_stream & 0x1F) << 8;
      data_stream++;

      elem->elementary_PID |= (u16)(*data_stream);
      data_stream++;

      desc_len  = (u16)(*data_stream & 0x0F) << 8;
      data_stream++;

      desc_len |= (u16)(*data_stream);
      data_stream++;

      stream_len -= (int)desc_len + 5;
      if (stream_len >= 0) {
         rc = PMT_DESC_read(data_stream, (int)desc_len, elem->desc,
                            desc, PMT_MAX_ELEM_DESC);
         if (rc < 0)
            goto ELEM_STREAM_read_error;
         else if (rc == 0)
            elem->desc_size = 0;
         else {
            elem->desc_size = (u8)rc;
            desc            = (u8 *)elem->desc[rc - 1];
            desc           += (u32)elem->desc[rc - 1]->desc_length + 2;
            data_stream    += desc_len;
         }

         elem++;         
         num_elem++;
         if ( (num_elem == (MPEG2_SYS_PIDS-2))&&(stream_len >= 5) ) {
            stream_len = 0;
/*
            if (PMT_MODULE_PRINT_DATA)
               printf("ELEM_STREAM_read: Need more than %d elementary streams\n", (MPEG2_SYS_PIDS-2));
*/
         }
      }
      else {
ELEM_STREAM_read_error:
         elem->desc_size = 0;
         stream_len      = 1;
      }
   }
   return ((stream_len == 0) ? num_elem : -1);
}

/*
static void PMT_display(PMT_SECTION *pmt, u16 ver, u16 prog_num)
{
   u8 i, j;

   printf("\nProgram Map Table\n"
          "   version     = %6u\n"
          "   program_num = 0x%04X\n", ver,
                                       prog_num);
   if (pmt->desc_size > 0) {
      printf("     PROGRAM DESCRIPTORS\n"
             "      TAG    LEN\n");
      for (i = 0; i < pmt->desc_size; i++) {
         printf("     0x%02hX   0x%02hX\n", (u16)pmt->desc[i]->desc_tag,
                                            (u16)pmt->desc[i]->data[0]);
      }
   }

   if (pmt->elem_size == 0)
      return;
      
   printf("     ELEMENTARY STREAMS\n"
          "     TYPE      PID\n");
   for (i = 0; i < pmt->elem_size; i++) {
      printf("     0x%02hX   0x%04hX\n", (u16)pmt->elem[i]->stream_type,
                                         pmt->elem[i]->elementary_PID);
      if (pmt->elem[i]->desc_size > 0) {
         printf("        ELEMENTARY STREAM %u DESCRIPTORS\n"
                "         TAG    LEN\n", (u32)i);
         for (j = 0; j < pmt->elem[i]->desc_size; j++) {
            printf("        0x%02hX   0x%02hX\n", (u16)pmt->elem[i]->desc[j]->desc_tag,
                                                  (u16)pmt->elem[i]->desc[j]->data[0]);
         }
      }
   }
}
*/

