/***************************************************************************
                          TS_PKT.h  -  description
                             -------------------
   The structures, and functions needed to deal with the transport stream
   packets.
                             -------------------
    begin                : Sun Jul 18 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __TS_PKT_H__
#define __TS_PKT_H__

#include "common/typedefs.h"

#define TS_PKT_SIZE                 188

#define TS_PKT_SYNC_BYTE           0x47

#define TS_PKT_PAT_ID            0x0000
#define TS_PKT_CAT_ID            0x0001
#define TS_PKT_ATSC_SI_ID        0x1FFB
#define TS_PKT_NUL_ID            0x1FFF
#define TS_PKT_RES_ID(ptr_to_TS_PKT) \
   ( (0x0001 < ptr_to_TS_PKT->PID)&&(ptr_to_TS_PKT->PID < 0x0010) )
#define TS_PKT_USR_ID(ptr_to_TS_PKT) \
   ( (0x000F < ptr_to_TS_PKT->PID)&&(ptr_to_TS_PKT->PID < 0x1FFF) )

#define TS_PKT_HAS_ADAPT_FIELD     0x20
#define TS_PKT_HAS_PAYLOAD         0x10

#define TS_PKT_CONT_CNT_MASK       0x0F

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
   u8  transport_error_indicator,      /*  1: Set when 1, or more uncorrected
                                                errors were detected in this
                                                packet. Can be cleared if the
                                                errors have been corrected. */
       payload_unit_start_indicator,   /*  1: Set when the payload contains
                                                the start of a PES, PSI, or
                                                PRIVATE_STREAM packet. Always
                                                cleared for NULL_PACKETs. */
       transport_priority,             /*  1: Set when this packet has a
                                                higher priority than other
                                                packets with equal PID, and a
                                                transport_priority of 0. */
       transport_scrambling_control;   /*  2: Indicates the scrambling mode of
                                                this packet's payload. HEADER
                                                and ADAPTATION_FIELD cannot be
                                                scrambled. NULL_PACKETs have a
                                                value of 0.
                   <0> not scrambled
                 <1-3> user defined
                   <1> ATSC stream is not scrambled, but access may be restricted
                   <2> ATSC stream is scrambled with even key
                   <3> ATSC stream is scrambled with odd key */
   u16 PID;                            /* 13: Indicates the type of data
                                                stored in the payload. Packets
                                                with a PID of 0x00, 0x01, or
                                                0x10-0x1FFE can contain a PCR
                                                (Program Clock Reference).
              <0x0000> Program Association Table
              <0x0001> Conditional Access Table
       <0x0002-0x000F> reserved
       <0x0010-0x1FFE> user defined (includes network_PID, program_map_PID, elementary_PID)
              <0x1FFF> NULL_PACKET
              <0x1FFB> ATSC System Information & Program Guide Data */
   u8  adaptation_field_control,       /*  2: Indicates what follows the
                                                HEADER of this packet.
                                                NULL_PACKETs have a value of
                                                '01'.
                <'00'> reserved for future use, can be ignored
                <'1X'> packet contains ADAPTATION_FIELD
                <'X1'> packet contains payload */
       continuity_counter;             /*  4: Counter that increments with each
                                                packet with equal PID. The
                                                counter does not increment if
                                                adaptation_field_control is
                                                '00', or '10'. A max of 2
                                                consecutive packets with equal
                                                PID and continuity_counter can
                                                appear in TRANSPORT_STREAM.
                                                Duplicate packets must be
                                                identical, except for the PCR
                                                (if present). The PCR must
                                                always be valid. */

   u32 payload_len32;
   const
   u8 *adapt_field_pos,               /*     Points to the location of the
                                                ADAPTATION FIELD, in the data
                                                stream, if present */
      *payload;                       /*     Points to the payload of this
                                                this packet, if it has any */
} TS_PKT;


/* Tries to find the first TS_PKT in data_stream.
      INPUT: data_stream - Points to the data to look through.
             data_len    - Number of bytes in data_stream. There should be at
                             least (TS_PKT_SIZE + 1) bytes to verify the
                             presence of a TS_PKT.
      RETURN: TS_PKT found     - A pointer to the TS_PKT_SYNC_BYTE, of the
                                   first TS_PKT.
              TS_PKT not found - NULL */
const u8 *TS_PKT_find(const u8 *data_stream, u32 data_len);

/* Reads the first TS_PKT, in data_stream, into ts.
      INPUT: data_stream - Points to the first byte, after the
                             TS_PKT_SYNC_BYTE, in the TS_PKT. This function
                             assumes there are at least (TS_PKT_SIZE - 1) 
                             bytes in data_stream.
      OUTPUT: ts - Points to the TS_PKT that will hold data taken from
                     data_stream.
      RETURN: Success - A pointer to the TS_PKT_SYNC_BYTE of the next TS_PKT
                          in data_stream (you must verify this yourself!).
              Error   - NULL */
const u8 *TS_PKT_read(const u8 *data_stream, TS_PKT *ts);

#ifdef __cplusplus
};
#endif

#endif
