/***************************************************************************
                          PIC_HEADER.c  -  description
                             -------------------
   Stuff needed for handling MPEG1 picture and gop headers.
                             -------------------
    begin                : Tue Sep 28 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifdef _M1V_TRACE_
# include <stdio.h>
#endif

#include "PIC_HEADER.h"

#ifdef _M1V_TRACE_
static void display_PIC_HEADER(PIC_HEADER1 *hdr);
#endif


/* Reads the gop header in data_stream.
      INPUT: data_stream - Points to the data stream holding the gop
                             header. This stream does not have to
                             contain the entire header. Repeated calls to
                             this function can be used to read the header
                             in parts.
             data_len    - Number of bytes in the data stream.
      OUTPUT: hdr     - Points to the PIC_HEADER1 that will hold the data.
              cur_pos - Used by MPEG1_VIDEO_MODULE.
              code    - Used by MPEG1_VIDEO_MODULE.
      RETURN: Number of bytes read. */
u32 GOP_HEADER1_read(const u8          *data_stream,
                           u32          data_len,
                           PIC_HEADER1 *hdr,
                           u32         *cur_pos,
                           u16         *code)
{
   switch (*cur_pos) {
   case 0:
      hdr->type &= 0xF8;
      hdr->type |= (*data_stream & 0x80) ? 0x04 : 0x00;
      if (data_len >= 4) {
         data_stream += 3;
         goto GOP_HEADER_read_FINISH;
      }

      *cur_pos = data_len;
      break;
   case 1:
   case 2:
      if (data_len <= (3 - *cur_pos)) {
         *cur_pos += data_len;
         break;
      }
      data_stream += 3 - *cur_pos;
   case 3:
GOP_HEADER_read_FINISH:
      hdr->type |= (*data_stream & 0x60) >> 5;
#ifdef _M1V_TRACE_
         printf("GOP_HEADER1> drop_frame = %c"
                           "  closed_gop = %c"
                           "  broken_link = %c\n",
                GOP_HEADER1_DROP_FRAME_FLAG(hdr) ?  'Y' : 'N',
                GOP_HEADER1_CLOSED_GOP_FLAG(hdr) ?  'Y' : 'N',
                GOP_HEADER1_BROKEN_LINK_FLAG(hdr) ? 'Y' : 'N');
#endif
      /* FINISHED READING THE GOP HEADER */
      data_len = 4 - *cur_pos;
      *cur_pos = 0;
      *code   |= 0x8000;   /* have to look for next start code */
      break;
   default:
#ifdef _M1V_TRACE_
      printf("GOP_HEADER1_read: ERROR illegal pos %u\n", *cur_pos);
#endif
      *cur_pos = 0;
      *code    = 0x0FFF;
      return 0;
   }

   return data_len;
}

/* Reads the picture header in data_stream.
      INPUT: data_stream - Points to the data stream holding the picture
                             header. This stream does not have to
                             contain the entire header. Repeated calls to
                             this function can be used to read the header
                             in parts.
             data_len    - Number of bytes in the data stream.
      OUTPUT: hdr     - Points to the PIC_HEADER1 that will hold the data.
              cur_pos - Used by MPEG1_VIDEO_MODULE.
              code    - Used by MPEG1_VIDEO_MODULE.
      RETURN: Number of bytes read. */
u32 PIC_HEADER1_read(const u8          *data_stream,
                           u32          data_len,
                           PIC_HEADER1 *hdr,
                           u32         *cur_pos,
                           u16         *code)
{
   const u8 *ptr = data_stream;

   switch (*cur_pos) {
   case 0:
      hdr->temporal_reference  = (u16)(*ptr) << 2;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 1;
         break;
      }
   case 1:
      hdr->temporal_reference |= (u16)(*ptr >> 6);
      hdr->type               &= 0x0F;
      hdr->type               |= (*ptr & 0x38) << 1;
      hdr->vbv_delay           = (u16)(*ptr & 0x07) << 13;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 2;
         break;
      }
   case 2:
      hdr->vbv_delay          |= (u16)(*ptr) << 5;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 3;
         break;
      }
   case 3:
      hdr->vbv_delay          |= (u16)(*ptr) >> 3;
      if (!(hdr->type & CODING_TYPE_P_FRAME))
         goto PIC_HEADER1_read_FINISH;

      hdr->f_code[0]  = (*ptr & 0x04) << 2;     /* full_pel_fwd_vector */
      hdr->f_code[0] |= (*ptr & 0x03) << 1;     /* fwd_f_code */
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 4;
         break;
      }
   case 4:
      hdr->f_code[0] |= *ptr >> 7;              /* fwd_f_code */
      if (PIC_HEADER1_CODING_TYPE(hdr) == CODING_TYPE_B_FRAME) {
         hdr->f_code[1]  = (*ptr & 0x40) >> 2;  /* full_pel_bck_vector */
         hdr->f_code[1] |= (*ptr & 0x38) >> 3;  /* bck_f_code */
      }
      
PIC_HEADER1_read_FINISH:
#ifdef _M1V_TRACE_
      display_PIC_HEADER(hdr);
#endif
      /* FINISHED READING THE PIC HEADER */
      ptr++;
      *cur_pos = 0;
      *code   |= 0x8000;   /* have to look for next start code */
      break;

   default:
#ifdef _M1V_TRACE_
      printf("PIC_HEADER1_read: ERROR illegal pos %u\n", *cur_pos);
#endif
      *cur_pos = 0;
      *code    = 0x0FFF;
      return 0;
   }

   return (u32)(ptr - data_stream);
}


#ifdef _M1V_TRACE_
static void display_PIC_HEADER(PIC_HEADER1 *hdr)
{
   const char *coding_type;

   switch (PIC_HEADER1_CODING_TYPE(hdr)) {
   case CODING_TYPE_I_FRAME:
      coding_type = "I Frame";
      break;
   case CODING_TYPE_P_FRAME:
      coding_type = "P Frame";
      break;
   case CODING_TYPE_B_FRAME:
      coding_type = "B Frame";
      break;
   case CODING_TYPE_D_FRAME:
      coding_type = "D Frame";
      break;
   default:
      coding_type = "Reserved";
   }

   printf("PIC_HEADER1> temporal_reference = %hu\n"
          "             coding_type        = %s\n"
          "             vbv_delay          = %hu\n",
          hdr->temporal_reference,
          coding_type,
          hdr->vbv_delay);

   if (hdr->type & CODING_TYPE_P_FRAME) {
      printf("             f_code             = "
             "0x%02hX", (u16)hdr->f_code[0]);
      if (hdr->type & CODING_TYPE_I_FRAME)
         printf(" 0x%02hX", (u16)hdr->f_code[1]);
      printf("\n\n");
   }
   else
      printf("\n");
}
#endif

