/***************************************************************************
                          SLICE.c  -  description
                             -------------------
   Stuff needed for handling MPEG1 slices.  This is where the video
   decoding process actually take place.
                             -------------------
    begin                : Tue Oct 5 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*#define _M1V_IDCT_TRACE_*/
/*#define _SHOW_ERRORS_*/
#ifdef _M1V_TRACE_
# ifndef _SHOW_ERRORS_
#  define _SHOW_ERRORS_
# endif
#endif

#ifdef _SHOW_ERRORS_
#include <stdio.h>
#endif
#include <string.h>

#include "SLICE.h"
#include "output_blocks.h"
#include "mpeg2_video/mpeg_tables.h"
#include "mpeg2_video/idct.h"
#include "mpeg2_video/motion_vectors.h"
#include "bitstream.h"


#define RESET_DCT_PRED(s_ptr) \
   s_ptr->dct_pred[0] = s_ptr->dct_pred[1] = s_ptr->dct_pred[2] = 0
#define RESET_PMV(s_ptr, bytes) \
   memset(s_ptr->PMV, 0, bytes)

#define USE_LOCAL_VARS
#ifdef USE_LOCAL_VARS
# define DATA_BYTE         data
# define BIT_SHIFT_BYTE    bit_shift
#else
# define DATA_BYTE         slice->data
# define BIT_SHIFT_BYTE    slice->bit_shift
#endif
#define  DATA_PTR          ptr
#define  END_DATA_PTR      end
#define  PARSER_STATE      *cur_pos
#define  STOP_PARSING      goto SLICE_read_END


#ifdef _M1V_TRACE_
static int frame_cnt = -1,
           mb_cnt    = -1,
           slice_cnt = -1,
           block_cnt = -1;
#endif

#ifdef _M1V_IDCT_TRACE_
static void show_QFS(s16 *QFS);
#endif

/* Processes a slice in data_stream.
      INPUT: data_stream - Points to the data stream holding the slice.
                             This stream does not have to contain the
                             entire slice. Repeated calls to this
                             function can be used to read the slice
                             in parts.
             data_len    - Number of bytes in the data stream.
      OUTPUT: seq_hdr - Points to the SEQ_HEADER1 of the mpeg1 stream.
              pic_hdr - Points to the PIC_HEADER1 of the slice.
              cur_pos - Used by MPEG1_VIDEO_MODULE.
              code    - Used by MPEG1_VIDEO_MODULE.
      RETURN: Number of bytes read. */
u32 SLICE1_read(const u8     *data_stream,
                      u32     data_len,
                      SLICE1 *slice,
                      u32    *cur_pos,
                      u16    *code)
{
   const u8 *ptr = data_stream,
            *end = data_stream + data_len;
   u32 tmp;
   s32 dct;
   u8  bit_shift_inc, run, level;
#ifdef USE_LOCAL_VARS
   u8  bit_shift, data;
#endif

#define SLICE_0                0
#define SLICE_1                1
#define SLICE_2                2
#define MACROBLOCK_0           3
#define MACROBLOCK_00          4
#define MACROBLOCK_1           5
#define MACROBLOCK_MODES_0     6
#define MACROBLOCK_MODES_1     7
#define MACROBLOCK_MODES_2     8
#define MACROBLOCK_2           9
#define MACROBLOCK_3          10
#define MACROBLOCK_4          11
#define MOTION_VECTOR_2       12
#define MOTION_VECTOR_3       13
#define MOTION_VECTOR_4       14
#define MOTION_VECTOR_5       15
#define MOTION_VECTOR_6       16
#define MACROBLOCK_6          17
#define MACROBLOCK_7          18
#define BLOCK_0               19
#define BLOCK_1               20
#define BLOCK_2               21
#define BLOCK_4               22
#define BLOCK_6               23
#define BLOCK_9               24
#define BLOCK_10              25
#define BLOCK_11              26
#define BLOCK_12              27
#define BLOCK_16              28
#define BLOCK_17              29
#define BLOCK_18              30
#define BLOCK_19              31
#define BLOCK_20              32
#define MACROBLOCK_8          33
#define MACROBLOCK_9          34
#define MACROBLOCK_10         35

START_AGAIN:
   if (*cur_pos) {
#ifdef USE_LOCAL_VARS
      bit_shift = slice->bit_shift;
      data      = slice->data;
#endif
      switch (*cur_pos) {
      case SLICE_1:
         goto SLICE_skip_extra_byte;
      case SLICE_2:
         goto SLICE_2_START;
      case MACROBLOCK_0:
         goto MACROBLOCK_read_addr_inc;
      case MACROBLOCK_00:
         goto MACROBLOCK_read_next_addr_inc;
      case MACROBLOCK_1:
         goto MACROBLOCK_1_START;
      case MACROBLOCK_MODES_0:
         goto MACROBLOCK_MODES_0_START;
      case MACROBLOCK_MODES_1:
         goto MACROBLOCK_MODES_P_FRAME;
      case MACROBLOCK_MODES_2:
         goto MACROBLOCK_MODES_B_FRAME;
      case MACROBLOCK_2:
         goto MACROBLOCK_2_START;
      case MACROBLOCK_3:
         goto MACROBLOCK_3_START;
      case MACROBLOCK_4:
         goto MACROBLOCK_read_MOTION_BCK;
      case MOTION_VECTOR_2:
         goto MOTION_VECTOR_read;
      case MOTION_VECTOR_3:
         goto MOTION_VECTOR_3_START;
      case MOTION_VECTOR_4:
         goto MOTION_VECTOR_4_START;
      case MOTION_VECTOR_5:
         goto MOTION_VECTOR_5_START;
      case MOTION_VECTOR_6:
         goto MOTION_VECTOR_6_START;
      case MACROBLOCK_6:
         goto MACROBLOCK_6_START;
      case MACROBLOCK_7:
         goto MACROBLOCK_7_START;
      case BLOCK_0:
         goto BLOCKS_read_start;
      case BLOCK_1:
         goto BLOCK_read;
      case BLOCK_2:
         goto BLOCK_2_START;
      case BLOCK_4:
         goto FIRST_INTRA_CHROMA_DCT_read;
      case BLOCK_6:
         goto BLOCK_6_START;
      case BLOCK_9:
         goto FIRST_DCT_COEF_read;
      case BLOCK_10:
         goto TABLE0_DCT_COEF_read;
      case BLOCK_11:
         goto BLOCK_11_START;
      case BLOCK_12:
         goto BLOCK_12_START;
      case BLOCK_16:
         goto BLOCK_16_START;
      case BLOCK_17:
         goto BLOCK_17_START;
      case BLOCK_18:
         goto BLOCK_18_START;
      case BLOCK_19:
         goto BLOCK_19_START;
      case BLOCK_20:
         goto BLOCK_read_done;
      case MACROBLOCK_8:
         goto MACROBLOCK_next;
      case MACROBLOCK_9:
         goto MACROBLOCK_9_START;
      case MACROBLOCK_10:
         goto MACROBLOCK_10_START;
      default:
         goto SLICE_read_ERROR;
      }
   }

   /* read SLICE */
/*   case SLICE_0: */
SLICE_start_new:
#ifdef _M1V_TRACE_
      if (frame_cnt < 0) {
         frame_cnt = slice_cnt = 0;
      }
      else if ((*code - 1) < slice->mb_row) {
         frame_cnt++;
         slice_cnt = 0;
      }
      else
         slice_cnt++;
#endif
      slice->mb_row       = *code - 1;
      slice->q_scale_code = *ptr >> 3;
      RESET_PMV(slice, (sizeof(s16)*8));  /* reset DCT and MV predictors */

      BIT_SHIFT_BYTE = 5;
      ALIGN8_DATA_BYTE(SLICE_1);
/*   case SLICE_1: */
SLICE_skip_extra_byte:
      GET_LSB;
      slice->r = DATA_BYTE & 0x80;  /* extra data flag */
      ALIGN_DATA_BYTE(SLICE_2, 1);
/*   case SLICE_2: */
SLICE_2_START:
      if (slice->r) {
         ALIGN8_DATA_BYTE(SLICE_1);
         goto SLICE_skip_extra_byte;
      }

   /* read MACROBLOCK */
#ifdef _M1V_TRACE_
      mb_cnt = (int)(slice->mb_row * slice->mb_wid) - 1;
#endif
      slice->mb_col = 0xFFFF;
/*   case MACROBLOCK_0: */
MACROBLOCK_read_addr_inc:
      slice->signed_level = 0;
/*   case MACROBLOCK_00: */
MACROBLOCK_read_next_addr_inc:
      GET_LSB;

      /* macroblock_addr_inc VLC: BYTE 1 */
      if (DATA_BYTE >= 0x06) {
         if (DATA_BYTE >= 0x80) {         /*-1-*/
            tmp = 1;
            bit_shift_inc = 1;
         }
         else if (DATA_BYTE >= 0x40) {    /* 01 */
            if (DATA_BYTE >= 0x60)        /*-011-*/
               tmp = 2;
            else                          /*-010-*/
               tmp = 3;
            bit_shift_inc = 3;
         }
         else if (DATA_BYTE >= 0x20) {    /* 001 */
            if (DATA_BYTE >= 0x30)        /*-0011-*/
               tmp = 4;
            else                          /*-0010-*/
               tmp = 5;
            bit_shift_inc = 4;
         }
         else if (DATA_BYTE >= 0x10) {    /* 0001 */
            if (DATA_BYTE >= 0x18)        /*-0001 1-*/
               tmp = 6;
            else                          /*-0001 0-*/
               tmp = 7;
            bit_shift_inc = 5;
         }
         else if (DATA_BYTE >= 0x0C) {    /* 0000 11 */
            if (DATA_BYTE >= 0x0E)        /*-0000 111-*/
               tmp = 8;
            else                          /*-0000 110-*/
               tmp = 9;
            bit_shift_inc = 7;
         }
         else {                           /* 0000 1011 -> 0000 0110 */
            tmp = 21 - (u32)DATA_BYTE;
            bit_shift_inc = 8;
         }

         /* finished reading VLC code */
         goto MACROBLOCK_read_addr_inc_done;
      }

      /* VLC code is longer than 8 bits, first 5 bits are 0 */
      ALIGN_DATA_BYTE(MACROBLOCK_1, 5);
/*   case MACROBLOCK_1: */
MACROBLOCK_1_START:
      GET_LSB;

      /* macroblock_addr_inc VLC: BYTE 2 */
      if (DATA_BYTE >= 0x90) {
         if (DATA_BYTE >= 0xA0)              /* 0000 0101 XX */
            tmp = 39 - (u32)(DATA_BYTE >> 3);
         else if (DATA_BYTE >= 0x98)         /*-0000 0100 11-*/
            tmp = 20;
         else                                /*-0000 0100 10-*/
            tmp = 21;
         bit_shift_inc = 5;
      }
      else if (DATA_BYTE >= 0x60) {          /* 0000 0100 0XX */
                                             /* 0000 0011 XXX */
         tmp = 57 - (u32)(DATA_BYTE >> 2);
         bit_shift_inc = 6;
      }
      else {
         if ( (0x20 <= DATA_BYTE)&&(DATA_BYTE <= 0x23) ) {
                                             /*-0000 0001 000 */
            slice->signed_level += 33; /* encountered VLC escape code */
         }
         else if ( (0x3C > DATA_BYTE)||(DATA_BYTE > 0x3F) ) {
#ifdef _SHOW_ERRORS_
            printf("\nInvalid mb_addr_inc VLC2 (0x%02hX)\n",
                   (u16)DATA_BYTE);
#endif
            goto SLICE_read_ERROR;
         }
         /* decode the next code */
         ALIGN_DATA_BYTE(MACROBLOCK_00, 6);
         goto MACROBLOCK_read_next_addr_inc;
      }

MACROBLOCK_read_addr_inc_done:
      tmp += (u32)slice->signed_level;
      if (tmp > 1) {
         if (PIC_HEADER1_CODING_TYPE(slice->pic_hdr) == CODING_TYPE_P_FRAME)
            RESET_PMV(slice, (sizeof(s16)*8));  /* reset DCT & MV predictors */
         else
            RESET_DCT_PRED(slice);  /* reset DCT predictors */

         if (slice->mb_col != 0xFFFF)
            skip_macroblocks1(slice, (tmp - 1));
         slice->mb_col += (u16)tmp;
      }
      else
         slice->mb_col++;
#ifdef _M1V_TRACE_
      mb_cnt += (int)tmp;
      printf("FRAME %d, SLICE %d, MACROBLOCK %d\n",
             frame_cnt, slice_cnt, mb_cnt);
      printf("  mb_col = %hu, mb_row = %hu\n"
             "  cur_col= %hu, cur_row= %hu\n"
             "  q_scale_code = 0x%02hX, VLC mb_addr_inc = %u\n",
             slice->mb_col,  slice->mb_row,
             slice->cur_col, slice->cur_row,
             (u16)slice->q_scale_code, tmp);
#endif
      ALIGN_DATA_BYTE(MACROBLOCK_MODES_0, bit_shift_inc);

   /* read MACROBLOCK_MODES */
/*   case MACROBLOCK_MODES_0: */
MACROBLOCK_MODES_0_START:
      GET_LSB;
      /* read mb_type VLC */
#ifdef _M1V_TRACE_
      printf("  VLC mb_type(");
#endif
      switch (PIC_HEADER1_CODING_TYPE(slice->pic_hdr)) {
      case CODING_TYPE_B_FRAME:
#ifdef _M1V_TRACE_
         printf("B");
#endif
         goto MACROBLOCK_MODES_B_FRAME;
      case CODING_TYPE_P_FRAME:
#ifdef _M1V_TRACE_
         printf("P");
#endif
         goto MACROBLOCK_MODES_P_FRAME;
      case CODING_TYPE_I_FRAME:
#ifdef _M1V_TRACE_
         printf("I");
#endif
         break;
      default:
#ifdef _SHOW_ERRORS_
         printf("\nInvalid picture_coding_type\n");
#endif
         goto SLICE_read_ERROR;
      }

      run = MB_INTRA_BIT;
      if (DATA_BYTE >= 0x80)        /*-1-*/
         bit_shift_inc = 1;
      else if (DATA_BYTE >= 0x40) { /*-01-*/
         run |= MB_QUANT_BIT;
         bit_shift_inc = 2;
      }
      else {
#ifdef _SHOW_ERRORS_
         printf("\nInvalid MACROBLOCK_MODES I frame VLC (0x%02hX)\n",
                (u16)DATA_BYTE);
#endif
         goto SLICE_read_ERROR;

      }
      goto MACROBLOCK_MODES_mb_type_done;

/*   case MACROBLOCK_MODES_1: */
MACROBLOCK_MODES_P_FRAME:
      if (DATA_BYTE >= 0x80) {      /*-1-*/
         run = MB_MOTION_FWD_BIT | MB_PATTERN_BIT;
         bit_shift_inc = 1;
      }
      else if (DATA_BYTE >= 0x40) { /*-01-*/
         run = MB_PATTERN_BIT;
         bit_shift_inc = 2;
      }
      else if (DATA_BYTE >= 0x20) { /*-001-*/
         run = MB_MOTION_FWD_BIT;
         bit_shift_inc = 3;
      }
      else if (DATA_BYTE >= 0x08) {
         if (DATA_BYTE >= 0x18)     /*-0001 1-*/
            run = MB_INTRA_BIT;
         else if (DATA_BYTE >= 0x10)/*-0001 0-*/
            run = MB_MOTION_FWD_BIT | MB_PATTERN_BIT | MB_QUANT_BIT;
         else                       /*-0000 1-*/
            run = MB_PATTERN_BIT | MB_QUANT_BIT;
         bit_shift_inc = 5;
      }
      else if (DATA_BYTE >= 0x04) { /*-0000 01-*/
         run = MB_INTRA_BIT | MB_QUANT_BIT;
         bit_shift_inc = 6;
      }
      else {
#ifdef _SHOW_ERRORS_
         printf("\nInvalid MACROBLOCK_MODES P frame VLC (0x%02hX)\n",
                (u16)DATA_BYTE);
#endif
         goto SLICE_read_ERROR;

      }

      if (!(run & (MB_INTRA_BIT|MB_MOTION_FWD_BIT))) {
         run |= MB_MISSING_MV_BIT;

         slice->vector[0][0] = 0;
         slice->vector[0][1] = 0;
         RESET_PMV(slice, (sizeof(s16)*4));  /* reset MV predictors */

         slice->mv_blk[0] = slice->ref_blkF;
         slice->mv_blk[1] = NULL;
      }
      goto MACROBLOCK_MODES_mb_type_done;

/*   case MACROBLOCK_MODES_2: */
MACROBLOCK_MODES_B_FRAME:
      if (DATA_BYTE >= 0x80) {      /* 1 */
         run = MB_MOTION_FWD_BIT | MB_MOTION_BCK_BIT;
         if (DATA_BYTE >= 0xC0)     /*-11-*/
            run |= MB_PATTERN_BIT;
         bit_shift_inc = 2;
      }
      else if (DATA_BYTE >= 0x40) { /* 01 */
         run = MB_MOTION_BCK_BIT;
         if (DATA_BYTE >= 0x60)     /*-011-*/
            run |= MB_PATTERN_BIT;
         bit_shift_inc = 3;
      }
      else if (DATA_BYTE >= 0x20) { /* 001 */
         run = MB_MOTION_FWD_BIT;
         if (DATA_BYTE >= 0x30)     /*-0011-*/
            run |= MB_PATTERN_BIT;
         bit_shift_inc = 4;
      }
      else if (DATA_BYTE >= 0x10) { /* 0001 */
         if (DATA_BYTE >= 0x18)     /*-0001 1-*/
            run = MB_INTRA_BIT;
         else {                     /*-0001 0-*/
            run = MB_MOTION_FWD_BIT | MB_MOTION_BCK_BIT |
                  MB_QUANT_BIT | MB_PATTERN_BIT;
         }
         bit_shift_inc = 5;
      }
      else if (DATA_BYTE >= 0x08) { /* 0000 1 */
         run = MB_QUANT_BIT | MB_PATTERN_BIT;
         if (DATA_BYTE >= 0x0C)     /*-0000 11-*/
            run |= MB_MOTION_FWD_BIT;
         else                       /*-0000 10-*/
            run |= MB_MOTION_BCK_BIT;
         bit_shift_inc = 6;
      }
      else if (DATA_BYTE >= 0x04) { /*-0000 01-*/
         run = MB_QUANT_BIT | MB_INTRA_BIT;
         bit_shift_inc = 6;
      }
      else {
#ifdef _SHOW_ERRORS_
         printf("\nInvalid MACROBLOCK_MODES B frame VLC (0x%02hX)\n",
                (u16)DATA_BYTE);
#endif
         goto SLICE_read_ERROR;

      }
MACROBLOCK_MODES_mb_type_done:
      slice->mb_flags = run;
#ifdef _M1V_TRACE_
      printf(") =%s%s%s%s%s\n",
             (MB_MOTION_FWD(slice)) ? " FWD" : "",
             (MB_MOTION_BCK(slice)) ? " BCK" : "",
             (MB_PATTERN(slice))    ? " CBP" : "",
             (MB_INTRA(slice))    ? " INTRA" : "",
             (MB_QUANT(slice))    ? " QUANT" : "");
#endif
      if (!(run & MB_INTRA_BIT)) {
         RESET_DCT_PRED(slice);  /* reset DCT predictors */
         if (!(run & MB_MISSING_MV_BIT))
            memset(slice->mv_blk, 0, (sizeof(u8 *)*2));
      }
      else if (PIC_HEADER1_CODING_TYPE(slice->pic_hdr) != CODING_TYPE_I_FRAME) {
         RESET_PMV(slice, (sizeof(s16)*4));  /* reset MV predictors */
      }
      ALIGN_DATA_BYTE(MACROBLOCK_2, bit_shift_inc);
   /* finished reading MACROBLOCK_MODES */

   /* read MACROBLOCK continue */
/*   case MACROBLOCK_2: */
MACROBLOCK_2_START:
      if (MB_QUANT(slice)) {
         GET_LSB;
         slice->q_scale_code = DATA_BYTE >> 3;
#ifdef _M1V_TRACE_
         printf("  q_scale_code = 0x%02hX\n", (u16)slice->q_scale_code);
#endif
         ALIGN_DATA_BYTE(MACROBLOCK_3, 5);
      }

/*   case MACROBLOCK_3: */
MACROBLOCK_3_START:
      if (MB_MOTION_FWD(slice)) {
         /* motion vector 0 (forward prediction) */
         slice->s = 0;
         goto MOTION_VECTOR_read_sel_1;
      }

/*   case MACROBLOCK_4: */
MACROBLOCK_read_MOTION_BCK:
      if (!MB_MOTION_BCK(slice))
         goto MACROBLOCK_6_START;

      /* motion vector 1 (backward prediction) */
      slice->s = 1;

/*   case MOTION_VECTOR_0: */
   /* read MOTION VECTOR (set slice->s to 0, 1) */
MOTION_VECTOR_read_sel_1:
      /* find the reference macroblocks for the motion vectors */
      slice->t = 0;
      if (slice->s)
         slice->mv_blk[1] = slice->ref_blkB;
      else
         slice->mv_blk[0] = slice->ref_blkF;

      /* slice->s, and slice->t set */
/*   case MOTION_VECTOR_2: */
MOTION_VECTOR_read:
      GET_LSB;

      /* motion_code VLC: BYTE 1 */
      if (DATA_BYTE >= 0x06) {
         if (DATA_BYTE >= 0x80) {      /*-1-*/
            run = 0;
            bit_shift_inc = 1;
         }
         else if (DATA_BYTE >= 0x40) { /*-01-*/
            run = 1;
            bit_shift_inc = 2;
         }
         else if (DATA_BYTE >= 0x20) { /*-001-*/
            run = 2;
            bit_shift_inc = 3;
         }
         else if (DATA_BYTE >= 0x10) { /*-0001-*/
            run = 3;
            bit_shift_inc = 4;
         }
         else if (DATA_BYTE >= 0x0C) { /*-0000 11-*/
            run = 4;
            bit_shift_inc = 6;
         }
         else {                        /* 0000 101 -> 0000 011 */
            run = 10 - (DATA_BYTE >> 1);
            bit_shift_inc = 7;
         }

         /* finished reading motion_code VLC */
         goto MOTION_VECTOR_read_motion_code_done;
      }

      /* motion_code is longer than 8 bits, first 5 bits are 0 */
      ALIGN_DATA_BYTE(MOTION_VECTOR_3, 5);
/*   case MOTION_VECTOR_3: */
MOTION_VECTOR_3_START:
      GET_LSB;

      /* motion_code VLC: BYTE 2 */
      if (DATA_BYTE >= 0x90) {
         if (DATA_BYTE >= 0xB0)        /*-0000 0101 1-*/
            run =  8;
         else if (DATA_BYTE >= 0xA0)   /*-0000 0101 0-*/
            run =  9;
         else                          /*-0000 0100 1-*/
            run = 10;
         bit_shift_inc = 4;
      }
      else if (DATA_BYTE >= 0x80) {
         if (DATA_BYTE >= 0x88)        /*-0000 0100 01-*/
            run = 11;
         else                          /*-0000 0100 00-*/
            run = 12;
         bit_shift_inc = 5;
      }
      else if (DATA_BYTE >= 0x60) {    /* 0000 0011 XX */
         run = 28 - (DATA_BYTE >> 3);
         bit_shift_inc = 5;
      }
      else {
#ifdef _SHOW_ERRORS_
         printf("\nInvalid motion_code VLC2 (0x%02hX)\n",
                (u16)DATA_BYTE);
#endif
         goto SLICE_read_ERROR;

      }
MOTION_VECTOR_read_motion_code_done:
      /* finished reading motion_code VLC */
      slice->signed_level = (s16)run;
      ALIGN_DATA_BYTE(MOTION_VECTOR_4, bit_shift_inc);
/*   case MOTION_VECTOR_4: */
MOTION_VECTOR_4_START:
#ifdef _M1V_TRACE_
      if (!slice->t) {
         printf("last PMV[%hu]=(%hd, %hd)\n",
                (u16)slice->s,
                slice->PMV[slice->s][0],
                slice->PMV[slice->s][1]);
      }
      printf("motion_code[s=%hu][t=%hu] =",
             (u16)slice->s, (u16)slice->t);
#endif
      if (slice->signed_level) {
         GET_LSB;
         /* get sign bit */
         slice->i = DATA_BYTE & 0x80;
#ifdef _M1V_TRACE_
         printf("%s%hd\n", (slice->i) ? " -" : " ", slice->signed_level);
#endif
         ALIGN_DATA_BYTE(MOTION_VECTOR_5, 1);
      }
      else {
#ifdef _M1V_TRACE_
         printf(" 0\n");
#endif
         slice->signed_level = slice->PMV[slice->s][slice->t];
         goto MOTION_VECTOR_calc_done;
      }

/*   case MOTION_VECTOR_5: */
MOTION_VECTOR_5_START:
      run = slice->pic_hdr->f_code[slice->s] & 0x0F;
      if ( (1 < run)&&(run < 8) ) {
         slice->cc = run - 1;
         slice->signed_level--;
         slice->signed_level <<= slice->cc;
         slice->signed_level++;

         GET_LSB;
         /* read motion_residual VLC */
         run                  = 9 - run;
         slice->signed_level += (s16)(DATA_BYTE >> run);
         if (slice->i)
            slice->signed_level = -slice->signed_level;
#ifdef _M1V_TRACE_
         printf("motion_residual[s=%hu][t=%hu] = %hu\n",
                (u16)slice->s, (u16)slice->t,
                (u16)(DATA_BYTE >> run));
#endif
         bit_shift_inc = 8 - run;
         ALIGN_DATA_BYTE(MOTION_VECTOR_6, bit_shift_inc);
      }
      else {
         slice->cc = 0;
         if (slice->i)
            slice->signed_level = -slice->signed_level;
      }

/*   case MOTION_VECTOR_6: */
MOTION_VECTOR_6_START:
      /* calculate motion vector */
      slice->signed_level += slice->PMV[slice->s][slice->t];

      if (slice->signed_level < (-16 << slice->cc))
         slice->signed_level += 32 << slice->cc;
      else if (slice->signed_level > ((16 << slice->cc) - 1))
         slice->signed_level -= 32 << slice->cc;

      slice->PMV[slice->s][slice->t] = slice->signed_level;

MOTION_VECTOR_calc_done:
      if (slice->pic_hdr->f_code[slice->s] & FULL_PIXEL_FLAG)
         slice->signed_level <<= 1;
      slice->vector[slice->s][slice->t] = slice->signed_level;
#ifdef _M1V_TRACE_
      if (slice->t) {
         printf("vector[%hu]=(%hd, %hd)  ",
                (u16)slice->s,
                slice->vector[slice->s][0],
                slice->vector[slice->s][1]);
         printf("PMV[%hu]=(%hd, %hd)\n",
                (u16)slice->s,
                slice->PMV[slice->s][0],
                slice->PMV[slice->s][1]);
      }
#endif
      if (!slice->t) {
         slice->t = 1;
         goto MOTION_VECTOR_read;         /* read vertical component of motion vector */
      }
      else if (!slice->s)
         goto MACROBLOCK_read_MOTION_BCK; /* read the backward prediction vectors */

/*   case MACROBLOCK_6: */
MACROBLOCK_6_START:
      /* slice->r = decoded cbp value */
      if (!MB_PATTERN(slice)) {
         slice->r = 0;
         goto BLOCKS_read_start;
      }

/*CODED_BLOCK_PATTERN_read:*/
      GET_LSB;
      /* coded block pattern VLC: BYTE 1 */
      if (DATA_BYTE >= 0x04) {
         if (DATA_BYTE >= 0xE0) {         /*-111-*/
            run = 60;
            bit_shift_inc = 3;
         }
         else if (DATA_BYTE >= 0xA0) {
            if (DATA_BYTE >= 0xD0)        /*-1101-*/
               run =  4;
            else if (DATA_BYTE >= 0xC0)   /*-1100-*/
               run =  8;
            else if (DATA_BYTE >= 0xB0)   /*-1011-*/
               run = 16;
            else                          /*-1010-*/
               run = 32;
            bit_shift_inc = 4;
         }
         else {
            if (DATA_BYTE >= 0x40) {      /* 1001 1 -> 0100 0 */
               tmp           = (u32)((DATA_BYTE >> 3) - 0x08);
               bit_shift_inc = 5;
            }
            else if (DATA_BYTE >= 0x30) { /* 0011 XX */
               tmp           = (u32)(DATA_BYTE >> 2);
               bit_shift_inc = 6;
            }
            else if (DATA_BYTE >= 0x20) { /* 0010 XXX */
               tmp           = (u32)(DATA_BYTE >> 1);
               bit_shift_inc = 7;
            }
            else {                        /* 000X XXXX */
               tmp           = (u32)(DATA_BYTE + (24 - 0x04));
               bit_shift_inc = 8;
            }
            run = CBP_TABLE[tmp];
         }

         /* finished reading coded block pattern VLC */
         slice->r = run;
         goto CODED_BLOCK_PATTERN_read_done;
      }

      /* coded block pattern is longer than 8 bits, first 6 bits are 0 */
      ALIGN_DATA_BYTE(MACROBLOCK_7, 6);
/*   case MACROBLOCK_7: */
MACROBLOCK_7_START:
      GET_LSB;
      /* slice->r = decoded cbp value */
      /* coded block pattern VLC: BYTE 2 */
      if (DATA_BYTE >= 0x20) {
         slice->r = CBP_TABLE[(DATA_BYTE >> 5) + (52 - 0x01)];
         bit_shift_inc = 3;
      }
      else {
#ifdef _SHOW_ERRORS_
         printf("\nInvalid coded block pattern VLC2 (0x%02hX)\n",
                (u16)DATA_BYTE);
#endif
         goto SLICE_read_ERROR;
      }
CODED_BLOCK_PATTERN_read_done:
#ifdef _M1V_TRACE_
      printf("  VLC coded_block_pattern = %hu\n", (u16)slice->r);
#endif
      ALIGN_DATA_BYTE(BLOCK_0, bit_shift_inc);

/*   case BLOCK_0: */
BLOCKS_read_start:
      /* slice->i = index of current BLOCK */
      slice->i = 0;

      /* slice->r, slice->s, and slice->i set */
      /* cbp,      num_BLKS      BLK_idx      */
/*   case BLOCK_1: */
BLOCK_read:
      MV_CLEAR_BLOCK(slice->QFS);
      MV_DONE;
      slice->mb_flags &= 0xFE;
BLOCK_read_no_clear:
#ifdef _M1V_TRACE_
      block_cnt++;
      printf("DCT VLC Table 0\n");
#endif
      if (slice->i < 4)
         slice->cc = 0;
      else if (slice->i & 0x01)
         slice->cc = 2;
      else
         slice->cc = 1;

      if ( (!(slice->r & (1 << (5-slice->i))))&&(!MB_INTRA(slice)) ) {
         /* skipping this BLOCK, go to the next */
#ifdef _M1V_TRACE_
         printf("SKIPPED BLOCK %hu\n", (u16)slice->i);
#endif
         output_block1(slice);
         slice->i++;
         if (slice->i >= 6)
            goto MACROBLOCK_next;
         goto BLOCK_read_no_clear;
      }

      /* slice->t = index to slice->QFS[] array */
      slice->t = 0;

      if (!MB_INTRA(slice))
         goto FIRST_DCT_COEF_read;
      else if (slice->cc)
         goto FIRST_INTRA_CHROMA_DCT_read;

/*   case BLOCK_2: */
BLOCK_2_START:
      GET_LSB;

      if (DATA_BYTE >= 0xFE) {
#ifdef _SHOW_ERRORS_
         printf("\nInvalid intra dct luma size VLC1 (0x%02hX)\n",
                (u16)DATA_BYTE);
#endif
         goto SLICE_read_ERROR;
      }

      /* intra_luma_dct_size VLC: BYTE 1 */
      if (DATA_BYTE < 0x80) {          /* 0 */
         if (DATA_BYTE < 0x40)         /*-00-*/
            run = 1;
         else                          /*-01-*/
            run = 2;
         bit_shift_inc = 2;
      }
      else if (DATA_BYTE < 0xE0) {
         if (DATA_BYTE < 0xA0)         /*-100-*/
            run = 0;
         else if (DATA_BYTE < 0xC0)    /*-101-*/
            run = 3;
         else                          /*-110-*/
            run = 4;
         bit_shift_inc = 3;
      }
      else if (DATA_BYTE < 0xF0) {     /*-1110-*/
         run = 5;
         bit_shift_inc = 4;
      }
      else if (DATA_BYTE < 0xF8) {     /*-1111 0-*/
         run = 6;
         bit_shift_inc = 5;
      }
      else if (DATA_BYTE < 0xFC) {     /*-1111 10-*/
         run = 7;
         bit_shift_inc = 6;
      }
      else {                           /*-1111 110-*/
         run = 8;
         bit_shift_inc = 7;
      }

      /* finished reading intra_luma_dct_size VLC */
      goto INTRA_DCT_SIZE_done;

/*   case BLOCK_4: */
FIRST_INTRA_CHROMA_DCT_read:
      GET_LSB;

      if (DATA_BYTE == 0xFF) {
#ifdef _SHOW_ERRORS_
         printf("\nInvalid intra dct chroma size VLC1 (0x%02hX)\n",
                (u16)DATA_BYTE);
#endif
         goto SLICE_read_ERROR;
      }

      /* intra_chroma_dct_size VLC: BYTE 1 */
      if (DATA_BYTE < 0xC0) {          /*-00-*/
                                       /*-01-*/
                                       /*-10-*/
         run = DATA_BYTE >> 6;
         bit_shift_inc = 2;
      }
      else if (DATA_BYTE < 0xE0) {     /*-110-*/
         run = 3;
         bit_shift_inc = 3;
      }
      else if (DATA_BYTE < 0xF0) {     /*-1110-*/
         run = 4;
         bit_shift_inc = 4;
      }
      else if (DATA_BYTE < 0xF8) {     /*-1111 0-*/
         run = 5;
         bit_shift_inc = 5;
      }
      else if (DATA_BYTE < 0xFC) {     /*-1111 10-*/
         run = 6;
         bit_shift_inc = 6;
      }
      else if (DATA_BYTE < 0xFE) {     /*-1111 110-*/
         run = 7;
         bit_shift_inc = 7;
      }
      else {                           /*-1111 1110-*/
         run = 8;
         bit_shift_inc = 8;
      }

      /* finished reading intra_chroma_dct_size VLC */
INTRA_DCT_SIZE_done:
      slice->dct_size = run;
      ALIGN_DATA_BYTE(BLOCK_6, bit_shift_inc);

      /* slice->r, slice->s, slice->t, slice->i, and slice->cc set */
      /* cbp,      num_BLKS, QFS_idx,  BLK_idx                     */
/*   case BLOCK_6: */
BLOCK_6_START:
#ifdef _M1V_TRACE_
      printf("  dct_dc_%s: dct_size = %hu, dct_diff = ",
             (slice->cc) ? "chroma" : "luma",
             (u16)slice->dct_size);
#endif
      if (!slice->dct_size) {
         bit_shift_inc       = 0;
         slice->signed_level = slice->dct_pred[slice->cc];
#ifdef _M1V_TRACE_
         printf("0");
#endif
      }
      else {
         GET_LSB;
         /* read dct_diff VLC: BYTE 1 */
         bit_shift_inc       = slice->dct_size;
         slice->signed_level = (s16)(DATA_BYTE >> (8 - slice->dct_size));
         if (slice->signed_level < (s16)(1 << (slice->dct_size - 1)))
            slice->signed_level -= (s16)(1 << slice->dct_size) - 1;
#ifdef _M1V_TRACE_
         printf("%hd", slice->signed_level);
#endif
         slice->signed_level       += slice->dct_pred[slice->cc];
         slice->dct_pred[slice->cc] = slice->signed_level;
      }

      /* intra inverse quantization */
      slice->signed_level <<= 3;
#ifdef _M1V_TRACE_
      printf("\nDC>  0 < 0, %hd\n", slice->signed_level);
#endif
      /* inverse scan location */
      run = 0;
      goto DCT_COEF_saturation;

/*   case BLOCK_9: */
FIRST_DCT_COEF_read:
      GET_LSB;

      if (DATA_BYTE >= 0x80) {   /*-1-*/
         run   = 0;
         level = 1;
         bit_shift_inc = 1;
         goto DCT_COEF_read_done;
      }
      goto TABLE0_DCT_COEF_read_cont;

/*   case BLOCK_10: */
TABLE0_DCT_COEF_read:
      GET_LSB;

      /* DCT_COEF VLC: BYTE 1 */
      if (DATA_BYTE >= 0x80) {      /* 1 */
         bit_shift_inc = 2;
         if (DATA_BYTE < 0xC0)      /*-10-*/
            goto BLOCK_read_done;
                                    /*-11-*/
         run   = 0;
         level = 1;

         /* finished reading DCT_COEF VLC */
         goto DCT_COEF_read_done;
      }

TABLE0_DCT_COEF_read_cont:
      if (DATA_BYTE >= 0x04) {
         if (DATA_BYTE >= 0x60) {         /*-011-*/
            bit_shift_inc = 3;
            run   = 1;
            level = 1;
         }
         else if (DATA_BYTE >= 0x40) {
            bit_shift_inc = 4;
            if (DATA_BYTE >= 0x50) {      /*-0101-*/
               run   = 2;
               level = 1;
            }
            else {                        /*-0100-*/
               run   = 0;
               level = 2;
            }
         }
         else if (DATA_BYTE >= 0x28) {
            bit_shift_inc = 5;
            if (DATA_BYTE >= 0x30) {
               level = 1;
               if (DATA_BYTE >= 0x38)     /*-0011 1-*/
                  run = 3;
               else                       /*-0011 0-*/
                  run = 4;
            }
            else {                        /*-0010 1-*/
               run   = 0;
               level = 3;
            }
         }
         else if (DATA_BYTE >= 0x20) {    /* 0010 0XXX */
            bit_shift_inc = 8;
            tmp   = (u32)DATA_BYTE - (32-8);
            run   = DCT_TABLE0[tmp][0];
            level = DCT_TABLE0[tmp][1];
         }
         else {
            if (DATA_BYTE >= 0x10) {      /* 0001 XX */
               bit_shift_inc = 6;
               tmp           = (u32)(DATA_BYTE >> 2) - 4;
            }
            else if (DATA_BYTE >= 0x08) { /* 0000 1XX */
               bit_shift_inc = 7;
               tmp           = (u32)(DATA_BYTE >> 1);    /* + (4-4);*/
            }
            else {                        /*-0000 01-*/
               bit_shift_inc = 6;
               goto DCT_COEF_read_escape;
            }
            run   = DCT_TABLE0[tmp][0];
            level = DCT_TABLE0[tmp][1];
         }
         /* finished reading DCT_COEF VLC */
         goto DCT_COEF_read_done;
      }
      /* DCT_COEF VLC is longer than 8 bits, first 6 bits are zero */
      ALIGN_DATA_BYTE(BLOCK_11, 6);
/*   case BLOCK_11: */
BLOCK_11_START:
      GET_LSB;

      /* DCT_COEF VLC: BYTE 2 */
      if (DATA_BYTE >= 0x10) {
         if (DATA_BYTE >= 0x80) {      /* 0000 001X XX */
            bit_shift_inc = 4;
            tmp           = (u32)(DATA_BYTE >> 4) + (16-8);
         }
         else if (DATA_BYTE >= 0x40) { /* 0000 0001 XXXX */
            bit_shift_inc = 6;
            tmp           = (u32)(DATA_BYTE >> 2) + (24-16);
         }
         else if (DATA_BYTE >= 0x20) { /* 0000 0000 1XXX X */
            bit_shift_inc = 7;
            tmp           = (u32)(DATA_BYTE >> 1) + (40-16);
         }
         else {                        /* 0000 0000 01XX XX */
            bit_shift_inc = 8;
            tmp           = (u32)DATA_BYTE + (56-16);
         }
         run   = DCT_TABLE0[tmp][0];
         level = DCT_TABLE0[tmp][1];

         /* finished reading DCT_COEF VLC */
         goto DCT_COEF_read_done;
      }
      /* DCT_COEF VLC is longer than 14 bits, first 10 bits are zero */
      ALIGN_DATA_BYTE(BLOCK_12, 4);
/*   case BLOCK_12: */
BLOCK_12_START:
      GET_LSB;

      /* DCT_COEF VLC: BYTE 3 */
      if (DATA_BYTE >= 0x80) {      /* 0000 0000 001X XXX */
         bit_shift_inc = 5;
         tmp           = (u32)(DATA_BYTE >> 3) + (72-16);
      }
      else if (DATA_BYTE >= 0x40) { /* 0000 0000 0001 XXXX */
         bit_shift_inc = 6;
         tmp           = (u32)(DATA_BYTE >> 2) + (88-16);
      }
      else {
#ifdef _SHOW_ERRORS_
         printf("\nInvalid DCT_COEF Table0 VLC\n");
#endif
         goto SLICE_read_ERROR;
      }
      run   = DCT_TABLE0[tmp][0];
      level = DCT_TABLE0[tmp][1];

      /* finished reading DCT_COEF VLC */

      /* slice->t */
      /* QFS_idx  */
DCT_COEF_read_done:
#ifdef _M1V_TRACE_
      printf("%2hu> %2hu <", (u16)slice->t, (u16)run);
#endif
      if ((slice->t + run) >= 64) {
         if (slice->t >= 64) {
#ifdef _SHOW_ERRORS_
            printf("\nDCT_COEF ERROR> QFS_idx(%hu) value too big\n",
                   (u16)slice->t);
#endif
            goto SLICE_read_ERROR;
         }
         slice->t  = 63;
      }
      else
         slice->t += run;

      slice->signed_level = (s16)level;
      ALIGN_DATA_BYTE(BLOCK_16, bit_shift_inc);
/*   case BLOCK_16: */
BLOCK_16_START:
      GET_LSB;

      if (DATA_BYTE & 0x80)
         slice->signed_level = -slice->signed_level;
#ifdef _M1V_TRACE_
      printf("%2hu, %hd\n", (u16)slice->t, slice->signed_level);
#endif
      bit_shift_inc = 1;
DCT_COEF_inv_quant:
      /* inverse scan location */
      run = SCAN_TABLE[0][slice->t];

      /* inverse quantization */
      dct = (s32)slice->signed_level << 1;
      if (!MB_INTRA(slice)) {
         if (dct > 0)
            dct++;
         else if (dct < 0)
            dct--;
         dct *= (s32)slice->seq_hdr->non_intra_q_matrix[run];
      }
      else
         dct *= (s32)slice->seq_hdr->intra_q_matrix[run];
      dct *= (s32)slice->q_scale_code;
      slice->signed_level = (s16)(dct / 16);

      /* mismatch control */
      if ( (slice->signed_level)&&(!(slice->signed_level & 1)) ) {
         if (slice->signed_level > 0)
            slice->signed_level--;
         else
            slice->signed_level++;
      }

      /* run,                 and slice->t */
      /* inverse_scan_location    QFS_idx  */
DCT_COEF_saturation:
      /* saturation */
      if (slice->signed_level > 2047)
         slice->signed_level =  2047;
      else if (slice->signed_level < -2048)
         slice->signed_level = -2048;

      slice->QFS[run] = slice->signed_level;
      slice->t++;
      ALIGN_DATA_BYTE(BLOCK_10, bit_shift_inc);
      goto TABLE0_DCT_COEF_read;

      /* slice->t */
      /* QFS_idx  */
DCT_COEF_read_escape:
      ALIGN_DATA_BYTE(BLOCK_17, bit_shift_inc);
/*   case BLOCK_17: */
BLOCK_17_START:
      GET_LSB;

      /* read 6 bit run code */
      run = DATA_BYTE >> 2;
#ifdef _M1V_TRACE_
      printf("%2hu> %2hu <", (u16)slice->t, (u16)run);
#endif
      if ((slice->t + run) >= 64) {
         if (slice->t >= 64) {
#ifdef _SHOW_ERRORS_
            printf("\nDCT_COEF ESC ERROR> QFS_idx(%hu) value too big\n",
                   (u16)slice->t);
#endif
            goto SLICE_read_ERROR;
         }
         slice->t  = 63;
      }
      else
         slice->t += run;

      /* read first 2 bits of level code */
      slice->signed_level = (s16)(DATA_BYTE & 0x03) << 6;
      ALIGN8_DATA_BYTE(BLOCK_18);
/*   case BLOCK_18: */
BLOCK_18_START:
      GET_LSB;

      /* read next 6 bits of level code */
      slice->signed_level |= (s16)DATA_BYTE >> 2;
      if (slice->signed_level == 128)
         slice->signed_level = -256;
      else if (slice->signed_level) {
         if (slice->signed_level > 128)
            slice->signed_level -= 256;
         goto DCT_COEF_read_level_code_done;
      }

      /* read next 2 bits of level code */
      slice->dct_size = DATA_BYTE << 6;
      ALIGN8_DATA_BYTE(BLOCK_19);
/*   case BLOCK_19: */
BLOCK_19_START:
      GET_LSB;

      /* read last 6 bits of level code */
      slice->dct_size     |= DATA_BYTE >> 2;
      slice->signed_level += (s16)slice->dct_size;
DCT_COEF_read_level_code_done:
#ifdef _M1V_TRACE_
      printf("%2hu, %hd  ESC\n", (u16)slice->t, slice->signed_level);
#endif
      bit_shift_inc = 6;
      goto DCT_COEF_inv_quant;

      /* slice->s, and slice->i set */
      /* num_BLKS,     BLK_idx      */
/*   case BLOCK_20: */
BLOCK_read_done:
#ifdef _M1V_TRACE_
      printf("EOB>%2d\n", block_cnt);
#endif

#ifdef _M1V_IDCT_TRACE_
/*if ( (slice->cur_col == 3)&&(slice->cur_row == 0) )*/
{
      printf("DCT coefficients for MB %d, block %hu\n", mb_cnt, (u16)slice->i);
      show_QFS(slice->QFS);
}
#endif
      idct(slice->QFS);
#ifdef _M1V_IDCT_TRACE_
/*if ( (slice->cur_col == 3)&&(slice->cur_row == 0) )*/
{
      printf("Recovered samples\n");
      show_QFS(slice->QFS);
}
#endif
      output_block1(slice);

      slice->i++;
      if (slice->i < 6)
         tmp = BLOCK_1;       /* BLOCK_read */
      else if ( (!slice->cur_col)&&(slice->cur_row == slice->mb_hgt) ) {
#ifdef _M1V_TRACE_
         printf("END PICTURE\n");
#endif
         *cur_pos = 0;
         *code    = 0x80FF;   /* finished with this picture */
         goto SLICE_read_END;
      }
      else
         tmp = MACROBLOCK_8;  /* MACROBLOCK_next */

      ALIGN_DATA_BYTE(tmp, bit_shift_inc);
      if (slice->i < 6)
         goto BLOCK_read;

/*   case MACROBLOCK_8: */
MACROBLOCK_next:
      GET_LSB;
#ifdef _M1V_TRACE_
         block_cnt = -1;
         printf("\n");
#endif
      /* no more MACROBLOCKs exist if there are 23 zero bits here */
      if (DATA_BYTE)
         goto MACROBLOCK_read_addr_inc;

      /* 8 zero bits, since end of last MACROBLOCK */
      slice->t      = DATA_BYTE;
      slice->i      = BIT_SHIFT_BYTE;
      slice->buf[0] = *ptr;
      ALIGN8_DATA_BYTE(MACROBLOCK_9);
/*   case MACROBLOCK_9: */
MACROBLOCK_9_START:
      GET_LSB;

      if (DATA_BYTE) {
         tmp = 1;
MACROBLOCK_next_start:
         slice->data      = slice->t;
         slice->bit_shift = slice->i;
         *cur_pos         = MACROBLOCK_0; /* MACROBLOCK_read_addr_inc */
         SLICE1_read(slice->buf, tmp, slice, cur_pos, code);
         if (*code == 0x0FFF)
            goto SLICE_read_END;
         goto START_AGAIN;
      }

      /* 16 zero bits, since end of last MACROBLOCK */
      slice->buf[1] = *ptr;
      ALIGN8_DATA_BYTE(MACROBLOCK_10);
/*   case MACROBLOCK_10: */
MACROBLOCK_10_START:
      GET_LSB;

      if (DATA_BYTE & 0xFE) {
         tmp = 2;
         goto MACROBLOCK_next_start;
      }

      /* 23 zero bits, since end of last MACROBLOCK (SLICE has ended) */
      if (!slice->buf[0])
         *cur_pos  = 1; /* found 1st byte of next start_code ? */
      else
         *cur_pos  = 0;
      if (!slice->buf[1])
         *cur_pos += 1; /* found 1st, or 2nd byte of next start_code ? */

      if ( (*cur_pos == 2)&&(*ptr == 1)&&((u32)(ptr+2) < data_len)&&
           (0x01 < *(ptr+1))&&(*(ptr+1) <= 0xAF) ) {
         *code = (u16)(*(ptr+1));
         ptr  += 2;
         goto SLICE_start_new;
      }
      *code |= 0x8000;

SLICE_read_END:
#ifdef USE_LOCAL_VARS
   slice->bit_shift = bit_shift;
   slice->data      = data;
#endif
   return (u32)(ptr - data_stream);

SLICE_read_ERROR:
#ifdef _M1V_TRACE_
   printf("SLICE1_read_ERROR:\n");
   printf("   frame=%d, slice=%d, mblock=%d, block=%d\n", frame_cnt,
                                                          slice_cnt,
                                                          mb_cnt,
                                                          block_cnt);
   if (block_cnt >= 0) {
      /* slice->r, slice->s, slice->t, slice->i, and slice->cc set */
      /* cbp,      num_BLKS, QFS_idx,  BLK_idx                     */
      printf("   BLK_idx=%hu, num_BLKS=%hu, cc=%hu, QFS_idx=%hu\n\n",
             (u16)slice->i,
             (u16)slice->s,
             (u16)slice->cc,
             (u16)slice->t);
   }
   else
      printf("\n");
#endif
   *cur_pos = 0;
   *code    = 0x0FFF;
   ptr++;
   return (u32)(ptr - data_stream);

#undef SLICE_0
#undef SLICE_1
#undef SLICE_2
#undef MACROBLOCK_0
#undef MACROBLOCK_00
#undef MACROBLOCK_1
#undef MACROBLOCK_MODES_0
#undef MACROBLOCK_MODES_1
#undef MACROBLOCK_MODES_2
#undef MACROBLOCK_2
#undef MACROBLOCK_3
#undef MACROBLOCK_4
#undef MOTION_VECTOR_2
#undef MOTION_VECTOR_3
#undef MOTION_VECTOR_4
#undef MOTION_VECTOR_5
#undef MOTION_VECTOR_6
#undef MACROBLOCK_6
#undef MACROBLOCK_7
#undef BLOCK_0
#undef BLOCK_1
#undef BLOCK_2
#undef BLOCK_4
#undef BLOCK_6
#undef BLOCK_9
#undef BLOCK_10
#undef BLOCK_11
#undef BLOCK_12
#undef BLOCK_16
#undef BLOCK_17
#undef BLOCK_18
#undef BLOCK_19
#undef BLOCK_20
#undef MACROBLOCK_8
#undef MACROBLOCK_9
#undef MACROBLOCK_10
}

#ifdef _M1V_IDCT_TRACE_
static void show_QFS(s16 *QFS)
{
   int tmp;

      for (tmp = 0; tmp < 8; tmp++) {
         printf("%6hd", QFS[tmp]);
      }
      printf("\n");
      for (; tmp < 16; tmp++) {
         printf("%6hd", QFS[tmp]);
      }
      printf("\n");
      for (; tmp < 24; tmp++) {
         printf("%6hd", QFS[tmp]);
      }
      printf("\n");
      for (; tmp < 32; tmp++) {
         printf("%6hd", QFS[tmp]);
      }
      printf("\n");
      for (; tmp < 40; tmp++) {
         printf("%6hd", QFS[tmp]);
      }
      printf("\n");
      for (; tmp < 48; tmp++) {
         printf("%6hd", QFS[tmp]);
      }
      printf("\n");
      for (; tmp < 56; tmp++) {
         printf("%6hd", QFS[tmp]);
      }
      printf("\n");
      for (; tmp < 64; tmp++) {
         printf("%6hd", QFS[tmp]);
      }
      printf("\n");
}
#endif

#undef RESET_DCT_PRED
#undef RESET_PMV
#undef DATA_BYTE
#undef BIT_SHIFT_BYTE
#undef DATA_PTR
#undef END_DATA_PTR
#undef PARSER_STATE
#undef STOP_PARSING

