/***************************************************************************
                          output_blocks.c  -  description
                             -------------------
   Functions used for placing decoded blocks into the frame buffer.
                             -------------------
    begin                : Tue Dec 14 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifdef _M2V_TRACE_
#include <stdio.h>
#endif
#include <stdlib.h>

#include "output_blocks.h"
#include "motion_vectors.h"


#define HALF_PIXEL_X             0x01
#define HALF_PIXEL_Y             0x02
#define HALF_PIXEL_MASK          0x03
#define HALF_PIXEL_COLOUR_SHIFT     2

#define DCT_TOP_FIELD         1
#define DCT_BOT_FIELD         2
#define DCT_FRAME             3
#define DCT_FRAME_TOP_LUMA    4
#define DCT_FRAME_BOT_LUMA    5


/* #define USE_SKIPPED_MACROBLOCK */
#ifdef USE_SKIPPED_MACROBLOCK
# if defined(__MSVC__)
static __inline void output_skipped_macroblock(SLICE *slice);
# else /* if defined(__GCC__) */
static inline void output_skipped_macroblock(SLICE *slice);
# endif
#endif


/* Place a decoded Block into the frame buffer.
      INPUT: slice - Points to the SLICE holding the decoded Block,
                       and the pointer to the frame buffer. */
void output_block(SLICE *slice)
{
   s16 *in0;
   u8  *out0;
   int  wid, dct_type;

   /* do a boundary check, before writing to frame buffer */
   if (slice->cur_row >= slice->mb_hgt)
      return;

   /* initialize some stuff */
   out0 = slice->pred_blk;
   wid  = (int)slice->frame_wid;
   if (slice->cc)
      wid >>= 1;
   if (PIC_HEADER_STRUCTURE(slice->pic_hdr) == STRUCTURE_FULL_FRAME) {
      if ( (!slice->cc)&&(MB_DCT_TYPE(slice)) ) {
         if (slice->i < 2)
            dct_type = DCT_FRAME_TOP_LUMA;
         else {
            dct_type = DCT_FRAME_BOT_LUMA;
            out0    += wid;
         }
         wid <<= 1;
      }
      else
         dct_type = DCT_FRAME;
   }
   else {
      if (PIC_HEADER_STRUCTURE(slice->pic_hdr) == STRUCTURE_BOT_FIELD) {
         dct_type = DCT_BOT_FIELD;
         out0    += wid;
      }
      else
         dct_type = DCT_TOP_FIELD;
      wid <<= 1;
   }
#ifdef _M2V_TRACE_
   printf("BLK: pred_blk = %u  out0 = %u  dct_type=%d\n",
          (u32)slice->pred_blk, (u32)out0, dct_type);
#endif

   if (!MB_INTRA(slice)) {    /* do the motion vector stuff */
      int  move_blk;
      s16  pred_blk[64], *in1;
      u16  in_wid, mv_wid, y_wid, idx;
      u8   shift,
           move_type = 0;
/*
      printf("MV_FORMAT=%hu  dct_type=%d  slice->i=%hu  4mvs=%s\n",
             (u16)slice->pred_type, dct_type, (u16)slice->i,
             ( (slice->mv_blk[2])||(slice->mv_blk[3]) ) ? "YES" : "NO");
*/
      in0    = pred_blk;
      in1    = NULL;
      in_wid = 8;
      mv_wid = (u16)wid;
      if (MV_FORMAT_FIELD(slice)) {
         if (dct_type == DCT_FRAME) {
            mv_wid <<= 1;
            in1      = in0 + 8;
            in_wid   = 16;
         }
         else if ( (slice->mv_blk[2])||(slice->mv_blk[3]) )
            move_type = 1;
      }
      else if (dct_type > DCT_FRAME) {
         mv_wid  >>= 1;
         move_type = 2;
      }

      if (!slice->i) {
         s16 mv_x, mv_y,
             min_x = (s16)slice->cur_col * -16,
             min_y = (s16)slice->cur_row * -16,
             max_x = (s16)(slice->seq_hdr->horiz_size - 16) + min_x,
             max_y = (s16) slice->seq_hdr->vert_size;

         if (MV_FORMAT_FIELD(slice)) {
            min_y >>=  1;
            max_y >>=  1;
            if ( (move_type)||(dct_type == DCT_FRAME) )
               max_y -=  8;
            else
               max_y -= 16;
         }
         else
            max_y  -= 16;
         max_y += min_y;
#ifdef _M2V_TRACE_
         printf("MV_WID: wid=%d  mv_wid=%hu  in_wid=%hu\n", wid, mv_wid, in_wid);
         printf("MV_LIMITS: min_x=%hd max_x=%hd  min_y=%hd max_y=%hd\n",
                min_x, max_x, min_y, max_y);
#endif

         /* use vector[]s to move mv_blk[]s to new locations */
         for (idx = 0; idx < 4; idx++) {
            if (slice->mv_blk[idx]) {
               if ( (slice->mv_blk[idx] == slice->ref_blkF)||
                    (slice->mv_blk[idx] == slice->ref_blkB) )
                  slice->move[idx] = 0;                  /* top field */
               else
                  slice->move[idx] = (int)mv_wid >> 2;   /* bottom field */
               shift = 0;
               
               mv_x = slice->vector[idx>>1][idx&1][0] >> 1;
               mv_y = slice->vector[idx>>1][idx&1][1] >> 1;
               if ( (idx >= 2)&&(move_type)&&(dct_type < DCT_FRAME) )
                  mv_y += 8; 

               if (slice->vector[idx>>1][idx&1][0] & 0x0001) {
                  shift |= HALF_PIXEL_X;
                  if (mv_x > (max_x-1)) {
                     mv_x = max_x-1;
                     goto MV_X_IS_GOOD;
                  }
               }
               else if (mv_x > max_x) {
                  mv_x = max_x;
                  goto MV_X_IS_GOOD;
               }
               if (mv_x < min_x)
                  mv_x = min_x;
MV_X_IS_GOOD:
               if (slice->vector[idx>>1][idx&1][1] & 0x0001) {
                  shift |= HALF_PIXEL_Y;
                  if (mv_y > (max_y-1)) {
                     mv_y = max_y-1;
                     goto MV_Y_IS_GOOD;
                  }
               }
               else if (mv_y > max_y) {
                  mv_y = max_y;
                  goto MV_Y_IS_GOOD;
               }
               if (mv_y < min_y)
                  mv_y = min_y;
MV_Y_IS_GOOD:
#ifdef _M2V_TRACE_
               printf("MV: horiz = %hd  vert = %hd  idx=%hd\n",
                      slice->vector[idx>>1][idx&1][0] >> 1,
                      slice->vector[idx>>1][idx&1][1] >> 1, idx);
               printf("          = %hd       = %hd\n", mv_x, mv_y);
               printf("MV: mv_blk = %u\n", (u32)slice->mv_blk[idx]);
#endif
               if ( (mv_x)||(mv_y) ) {
                  move_blk  = (int)mv_x + ((int)mv_y * (int)mv_wid);
                  slice->mv_blk[idx] += move_blk;

                  /* change from (DIV 2) to (/ 2) */
                  if ( (mv_x < 0)&&(shift & HALF_PIXEL_X) )
                     mv_x++;
                  if ( (mv_y < 0)&&(shift & HALF_PIXEL_Y) )
                     mv_y++;
                  
                  move_blk -= (int) mv_x >> 1;
                  move_blk -= (int)(mv_y >> 1) * (int)(mv_wid >> 1);
                  slice->move[idx]   += move_blk;
#ifdef _M2V_TRACE_
                  printf("           = %u\n", (u32)slice->mv_blk[idx]);
#endif
                  if (mv_x & 0x0001)
                     shift |= HALF_PIXEL_X << HALF_PIXEL_COLOUR_SHIFT;
                  if (mv_y & 0x0001)
                     shift |= HALF_PIXEL_Y << HALF_PIXEL_COLOUR_SHIFT;
               }
               slice->half_pixel[idx] = shift;
            }
         }
      }

      idx = 0;
      switch (slice->i) {
      case 0:
         move_blk  =  8;
         break;
      case 1:
         if (move_type == 1)
            move_blk  =  0;
         else if (move_type == 2)
            move_blk  = -8;
         else {
            move_blk  = wid << 3;
            move_blk -=  8;
         }
         break;
      case 2:
         move_blk  =  8;
         if (move_type == 1)
            idx = 2;
         else if (move_type == 2) {
            if (slice->mv_blk[0]) {
               slice->mv_blk[0] += mv_wid;
               slice->move[0]   += mv_wid;
            }
            if (slice->mv_blk[1]) {
               slice->mv_blk[1] += mv_wid;
               slice->move[1]   += mv_wid;
            }
         }
         break;
      case 3:
         move_blk  = wid;
         if (move_type) {
            if (dct_type < DCT_FRAME)
               move_blk *= 12 * (int)slice->cur_row;
            else
               move_blk *=  6 * (int)slice->cur_row;
         }
         else
            move_blk *=(12 * (int)slice->cur_row) + 8;
         move_blk += (1 + (int)slice->cur_col) << 3;
         move_blk  = (int)((slice->luma_size * 5) >> 2) - move_blk;

         if (move_type == 1) {
            if (slice->mv_blk[0])
               slice->mv_blk[0] += move_blk - slice->move[0];
            if (slice->mv_blk[1])
               slice->mv_blk[1] += move_blk - slice->move[1];
            idx = 2;
         }
         break;
      case 4:
         move_blk  = (int)slice->luma_size / -4;
         if (move_type)
            in1 = in0 + 32;

         slice->half_pixel[0] >>= HALF_PIXEL_COLOUR_SHIFT;
         slice->half_pixel[1] >>= HALF_PIXEL_COLOUR_SHIFT;
         slice->half_pixel[2] >>= HALF_PIXEL_COLOUR_SHIFT;
         slice->half_pixel[3] >>= HALF_PIXEL_COLOUR_SHIFT;
         break;
      case 5:
         move_blk  =  0;
         if (move_type)
            in1 = in0 + 32;
      }
      shift = ( (slice->mv_blk[0])&&(slice->mv_blk[1]) ) ? 1 : 0;

      y_wid = mv_wid;
      if (move_type == 2)
         mv_wid <<= 1;
#ifdef _M2V_TRACE_
      printf("    move_type=%hu\n", (u16)move_type);
      printf("MV: mv_blk0=%u mv_blk1=%u mv_blk2=%u mv_blk3=%u\n",
             (u32)slice->mv_blk[0], (u32)slice->mv_blk[1],
             (u32)slice->mv_blk[2], (u32)slice->mv_blk[3]);
#endif

      /* get the predictor block */
      MV_CLEAR_BLOCK(in0);
      if (in1) {
         do {
            if (slice->mv_blk[idx]) {
               move_type = slice->half_pixel[idx] & HALF_PIXEL_MASK;
#ifdef _M2V_TRACE_
               printf("    half%hu=%hu  ", idx, (u16)move_type);
#else
               if (!move_type) {
                  MV_FIELD_PRED___(in0, in_wid, slice->mv_blk[idx],
                                   mv_wid);
               }
               else if (move_type == HALF_PIXEL_X) {
                  MV_FIELD_PRED_X_(in0, in_wid, slice->mv_blk[idx],
                                   mv_wid);
               }
               else if (move_type == HALF_PIXEL_Y) {
                  MV_FIELD_PRED__Y(in0, in_wid, slice->mv_blk[idx],
                                   mv_wid, y_wid);
               }
               else {
                  MV_FIELD_PRED_XY(in0, in_wid, slice->mv_blk[idx],
                                   mv_wid, y_wid);
               } 
#endif
               move_type = slice->half_pixel[idx+2] & HALF_PIXEL_MASK;
#ifdef _M2V_TRACE_
               printf("half%hu=%hu  mv_wid=%hu  y_wid=%hu  in_wid=%hu\n",
                      idx+2, (u16)move_type, mv_wid, y_wid, in_wid);
#else
               if (!move_type) {
                  MV_FIELD_PRED___(in1, in_wid, slice->mv_blk[idx+2],
                                   mv_wid);
               }
               else if (move_type == HALF_PIXEL_X) {
                  MV_FIELD_PRED_X_(in1, in_wid, slice->mv_blk[idx+2],
                                   mv_wid);
               }
               else if (move_type == HALF_PIXEL_Y) {
                  MV_FIELD_PRED__Y(in1, in_wid, slice->mv_blk[idx+2],
                                   mv_wid, y_wid);
               }
               else {
                  MV_FIELD_PRED_XY(in1, in_wid, slice->mv_blk[idx+2],
                                   mv_wid, y_wid);
               } 
#endif
               if (slice->i == 3) {
                  slice->mv_blk[idx]   -= slice->move[idx];
                  slice->mv_blk[idx+2] -= slice->move[idx+2];
               }
               slice->mv_blk[idx]   += move_blk;
               slice->mv_blk[idx+2] += move_blk;
            }
            idx++;
         } while (idx&1);
      }
      else {
         do {
            if (slice->mv_blk[idx]) {
               move_type = slice->half_pixel[idx] & HALF_PIXEL_MASK;
#ifdef _M2V_TRACE_
               printf("    half%hu=%hu  mv_wid=%hu  y_wid=%hu\n",
                      idx, (u16)move_type, mv_wid, y_wid);
                      
               if (slice->mv_blk[idx] < (u8 *)10000000)
                  printf("MV_ERROR: mv_blk[%d]=%u < 1buf __%hu\n", idx, (u32)slice->mv_blk[idx], (u16)slice->half_pixel[idx]);
               else if (slice->mv_blk[idx] < (u8 *)20000000) {
                  if ((slice->mv_blk[idx]+(mv_wid*7)+64) > (u8 *)(10000000+slice->frame_size))
                     printf("MV_ERROR: mv_blk[%d]=%u >= 1buf+%u __%hu\n", idx, (u32)slice->mv_blk[idx], slice->frame_size, (u16)slice->half_pixel[idx]);
               }
               else if ((slice->mv_blk[idx]+(mv_wid*7)+64) > (u8 *)(20000000+slice->frame_size))
                  printf("MV_ERROR: mv_blk[%d]=%u >= 2buf+%u __%hu\n", idx, (u32)slice->mv_blk[idx], slice->frame_size, (u16)slice->half_pixel[idx]);
#else
               if (!move_type) { 
                  MV_PRED___(in0, slice->mv_blk[idx],
                             mv_wid);
               }
               else if (move_type == HALF_PIXEL_X) {
                  MV_PRED_X_(in0, slice->mv_blk[idx],
                             mv_wid);
               }
               else if (move_type == HALF_PIXEL_Y) {
                  MV_PRED__Y(in0, slice->mv_blk[idx],
                             mv_wid, y_wid);
               }
               else {
                  MV_PRED_XY(in0, slice->mv_blk[idx],
                             mv_wid, y_wid);
               } 
#endif
               if (slice->i == 3)
                  slice->mv_blk[idx] -= slice->move[idx];
               slice->mv_blk[idx] += move_blk;
            }
            idx++;
         } while (idx&1);
      }

#ifdef _M2V_TRACE_
      printf("           =%u        =%u        =%u        =%u\n",
             (u32)slice->mv_blk[0], (u32)slice->mv_blk[1],
             (u32)slice->mv_blk[2], (u32)slice->mv_blk[3]);
#endif
      in0 = slice->QFS;
      in1 = pred_blk;
#ifdef _M2V_TRACE_
      if (out0 < slice->pred_frame)
         printf("SAT_ERROR: out0 < pred_frame  wid=%d\n", wid);
      else if ((out0+(wid*7)+64) > (slice->pred_frame+slice->frame_size))
         printf("SAT_ERROR: out0 > pred_frame+%u  wid=%d\n", slice->frame_size, wid);
#else
      /* put the final block into the frame */
      MV_ADD_PRED(in0, in1, out0, wid, shift);
#endif
   }
   else {
      in0 = slice->QFS;
#ifdef _M2V_TRACE_
      if (out0 < slice->pred_frame)
         printf("SAT_ERROR: out0 < pred_frame  wid=%d\n", wid);
      else if ((out0+(wid*7)+64) > (slice->pred_frame+slice->frame_size))
         printf("SAT_ERROR: out0 > pred_frame+%u  wid=%d\n", slice->frame_size, wid);
#else
      /* put the final block into the frame */
      MV_SATURATION(in0, out0, wid);
#endif
   }

   /* move frame pointers into correct position */
   switch (slice->i) {
   case 0:
      wid   =  8;
      break;
   case 1:
      if (dct_type <= DCT_FRAME) {
         wid <<=  3;
         wid  -=  8;
      }
      else
         wid   = -8;
      break;
   case 2:
      wid   =  8;
      break;
   case 3:
      if (dct_type <= DCT_FRAME)
         wid  *=(12 * (int)slice->cur_row) + 8;
      else
         wid  *=  6 * (int)slice->cur_row;
      wid  += (1 + (int)slice->cur_col) << 3;
      wid   = (int)((slice->luma_size * 5) >> 2) - wid;
      break;
   case 4:
      wid   = (int)slice->luma_size / -4;
      break;
   case 5:
      slice->ref_blkF += 16;
      slice->ref_blkB += 16;
      if ((slice->cur_col+1) == slice->mb_wid) {
         if (dct_type < DCT_FRAME) {
            dct_type = 31 * wid;
            wid *=(24 * (int)slice->cur_row) + 31;
         }
         else {
            dct_type = 30 * wid;
            wid *=(24 * (int)slice->cur_row) + 30;
         }
         dct_type += (int)slice->frame_wid - (int)slice->seq_hdr->horiz_size;
         wid      += (int)slice->frame_wid - (int)slice->seq_hdr->horiz_size;
         slice->ref_blkF += dct_type;
         slice->ref_blkB += dct_type;
      }
      else {
         dct_type = 0;
         wid *= 24 * (int)slice->cur_row;
      }
      wid  += (2 + (int)slice->cur_col) << 3;
      wid  -= (int)slice->luma_size;

      if (dct_type) {
         slice->cur_col = 0;
         slice->cur_row++;
      }
      else
         slice->cur_col++;
#ifdef _M2V_TRACE_
      printf("MV_REF: blkF=%u  blkB=%u\n", (u32)slice->ref_blkF, (u32)slice->ref_blkB);
#endif
   }

   slice->pred_blk += wid;
#ifdef _M2V_TRACE_
   printf("              = %u\n", (u32)slice->pred_blk);
#endif
}

/* Place skipped Macro-Blocks into the frame buffer.
      INPUT: slice - Points to the SLICE holding the pointer to the
                       frame buffer.
             cnt   - Number of Macro-Blocks that were skipped. */
void skip_macroblocks(SLICE *slice, u32 cnt)
{
   u32 inc;
   u8  mb_flags;

#ifndef USE_SKIPPED_MACROBLOCK
   MV_CLEAR_BLOCK(slice->QFS);
#endif
   if (PIC_HEADER_STRUCTURE(slice->pic_hdr) == STRUCTURE_FULL_FRAME) {
      slice->pred_type = PRED_TYPE_FRAME_BASED;
      inc = 0;
   }
   else {
      slice->pred_type = PRED_TYPE_FIELD_BASED;
      inc = (PIC_HEADER_STRUCTURE(slice->pic_hdr) == STRUCTURE_BOT_FIELD) ?
            slice->frame_wid : 0;
   }

   if (PIC_HEADER_CODING_TYPE(slice->pic_hdr) == CODING_TYPE_B_FRAME)
      mb_flags = slice->mb_flags;
   else {
      slice->vector[0][0][0] = 0;
      slice->vector[0][0][1] = 0;
      slice->vector[1][0][0] = 0;
      slice->vector[1][0][1] = 0;
      mb_flags = MB_MOTION_FWD_BIT;
   }
   
   slice->mb_flags  = 0;
   slice->mv_blk[0] = NULL;
   slice->mv_blk[1] = NULL;
   slice->mv_blk[2] = NULL;
   slice->mv_blk[3] = NULL;
   do {
      if (mb_flags & MB_MOTION_FWD_BIT)
         slice->mv_blk[0] = slice->ref_blkF + inc;
      if (mb_flags & MB_MOTION_BCK_BIT)
         slice->mv_blk[1] = slice->ref_blkB + inc;

#ifdef _M2V_TRACE_
      printf("SKIPPED MACROBLOCK\n"
             "  cur_col= %hu, cur_row= %hu\n",
             slice->cur_col, slice->cur_row);
#endif
#ifdef USE_SKIPPED_MACROBLOCK
      output_skipped_macroblock(slice);
#else
      for (slice->i = 0; slice->i < 6; slice->i++) {
         if (slice->i < 4)
            slice->cc = 0;
         else if (slice->i & 0x01)
            slice->cc = 2;
         else
            slice->cc = 1;
# ifdef _M2V_TRACE_
         printf("SKIPPED BLOCK %hu\n", (u16)slice->i);
# endif
         output_block(slice);
      }
#endif
      cnt--;
   } while (cnt);
}


#ifdef USE_SKIPPED_MACROBLOCK
# if defined(__MSVC__)
static __inline void output_skipped_macroblock(SLICE *slice)
# else /* if defined(__GCC__) */
static inline void output_skipped_macroblock(SLICE *slice)
# endif
{
   s16 *in0;
   u8  *out0;
   int  wid, dct_type, move_blk, y_wid, horiz_size;
   s16  i;
   u8   idx, shift, half_pix;

   /* do a boundary check, before writing to frame buffer */
   if (slice->cur_row >= slice->mb_hgt)
      return;

   /* initialize some stuff */
   i          = 0;
   horiz_size = (int)slice->frame_wid;
   if (PIC_HEADER_STRUCTURE(slice->pic_hdr) == STRUCTURE_FULL_FRAME)
      dct_type = DCT_FRAME;
   else {
      if (PIC_HEADER_STRUCTURE(slice->pic_hdr) == STRUCTURE_BOT_FIELD)
         dct_type = DCT_BOT_FIELD;
      else
         dct_type = DCT_TOP_FIELD;
      horiz_size <<= 1;
   }

START_LOOP:
   out0 = slice->pred_blk;
   wid  = horiz_size;
   if (i > 3)
      wid >>= 1;
   if (dct_type == DCT_BOT_FIELD)
      out0 += wid >> 1;

      if (!i) {
         s16 mv_x, mv_y,
             min_x = (s16)slice->cur_col * -16,
             min_y = (s16)slice->cur_row * -16,
             max_x = (s16)(slice->seq_hdr->horiz_size - 16) + min_x,
             max_y = (s16) slice->seq_hdr->vert_size;

         if (MV_FORMAT_FIELD(slice)) {
            min_y >>= 1;
            max_y >>= 1;
         }
         max_y += min_y - 16;

         /* use vector[]s to move mv_blk[]s to new locations */
         for (idx = 0; idx < 2; idx++) {
            if (slice->mv_blk[idx]) {
               if (dct_type != DCT_BOT_FIELD)
                  slice->move[idx] = 0;         /* top field */
               else
                  slice->move[idx] = wid >> 2;  /* bottom field */
               shift = 0;

               mv_x = slice->vector[0][idx][0] >> 1;
               mv_y = slice->vector[0][idx][1] >> 1;

               if (slice->vector[0][idx][0] & 0x0001) {
                  shift |= HALF_PIXEL_X;
                  if (mv_x > (max_x-1)) {
                     mv_x = max_x-1;
                     goto MV_X_IS_GOOD;
                  }
               }
               else if (mv_x > max_x) {
                  mv_x = max_x;
                  goto MV_X_IS_GOOD;
               }
               if (mv_x < min_x)
                  mv_x = min_x;
MV_X_IS_GOOD:
               if (slice->vector[0][idx][1] & 0x0001) {
                  shift |= HALF_PIXEL_Y;
                  if (mv_y > (max_y-1)) {
                     mv_y = max_y-1;
                     goto MV_Y_IS_GOOD;
                  }
               }
               else if (mv_y > max_y) {
                  mv_y = max_y;
                  goto MV_Y_IS_GOOD;
               }
               if (mv_y < min_y)
                  mv_y = min_y;
MV_Y_IS_GOOD:
               if ( (mv_x)||(mv_y) ) {
                  move_blk  = (int)mv_x + ((int)mv_y * wid);
                  slice->mv_blk[idx] += move_blk;

                  /* change from (DIV 2) to (/ 2) */
                  if ( (mv_x < 0)&&(shift & HALF_PIXEL_X) )
                     mv_x++;
                  if ( (mv_y < 0)&&(shift & HALF_PIXEL_Y) )
                     mv_y++;

                  move_blk -= (int) mv_x >> 1;
                  move_blk -= (int)(mv_y >> 1) * (wid >> 1);
                  slice->move[idx]   += move_blk;
                  if (mv_x & 0x0001)
                     shift |= HALF_PIXEL_X << HALF_PIXEL_COLOUR_SHIFT;
                  if (mv_y & 0x0001)
                     shift |= HALF_PIXEL_Y << HALF_PIXEL_COLOUR_SHIFT;
               }
               slice->half_pixel[idx] = shift;
            }
         }
      }

      idx = 0;
      switch (i) {
      case 0:
         move_blk  =  8;
         break;
      case 1:
         move_blk  = wid << 3;
         move_blk -=  8;
         break;
      case 2:
         move_blk  =  8;
         break;
      case 3:
         move_blk  = wid;
         move_blk *=(12 * (int)slice->cur_row) + 8;
         move_blk += (1 + (int)slice->cur_col) << 3;
         move_blk  = (int)((slice->luma_size * 5) >> 2) - move_blk;
         break;
      case 4:
         move_blk  = (int)slice->luma_size / -4;

         slice->half_pixel[0] >>= HALF_PIXEL_COLOUR_SHIFT;
         slice->half_pixel[1] >>= HALF_PIXEL_COLOUR_SHIFT;
         break;
      case 5:
         move_blk  =  0;
      }
      shift = ( (slice->mv_blk[0])&&(slice->mv_blk[1]) ) ? 1 : 0;

      /* get the predictor block */
      y_wid = wid;
      in0 = slice->QFS;
      MV_CLEAR_BLOCK(in0);
      do {
         if (slice->mv_blk[idx]) {
            half_pix = slice->half_pixel[idx] & HALF_PIXEL_MASK;
            if (!half_pix) {
               MV_PRED___(in0, slice->mv_blk[idx], wid, MVP_LOOP1);
            }
            else if (half_pix == HALF_PIXEL_X) {
               MV_PRED_X_(in0, slice->mv_blk[idx], wid, MVP_X_LOOP1);
            }
            else if (half_pix == HALF_PIXEL_Y) {
               MV_PRED__Y(in0, slice->mv_blk[idx], wid, y_wid, MVP_Y_LOOP1);
            }
            else {
               MV_PRED_XY(in0, slice->mv_blk[idx], wid, y_wid, MVP_XY_LOOP1);
            }

            if (i == 3)
               slice->mv_blk[idx] -= slice->move[idx];
            slice->mv_blk[idx] += move_blk;
         }
         idx++;
      } while (idx&1);

      /* put the final block into the frame */
      MV_COPY_PRED(in0, out0, wid, shift);

   /* move frame pointers into correct position */
   switch (i) {
   case 0:
      wid   =  8;
      break;
   case 1:
      if (dct_type <= DCT_FRAME) {
         wid <<=  3;
         wid  -=  8;
      }
      else
         wid   = -8;
      break;
   case 2:
      wid   =  8;
      break;
   case 3:
      if (dct_type <= DCT_FRAME)
         wid  *=(12 * (int)slice->cur_row) + 8;
      else
         wid  *=  6 * (int)slice->cur_row;
      wid  += (1 + (int)slice->cur_col) << 3;
      wid   = (int)((slice->luma_size * 5) >> 2) - wid;
      break;
   case 4:
      wid   = (int)slice->luma_size / -4;
      break;
   case 5:
      slice->ref_blkF += 16;
      slice->ref_blkB += 16;
      if ((slice->cur_col+1) == slice->mb_wid) {
         if (dct_type < DCT_FRAME) {
            dct_type = 31 * wid;
            wid *=(24 * (int)slice->cur_row) + 31;
         }
         else {
            dct_type = 30 * wid;
            wid *=(24 * (int)slice->cur_row) + 30;
         }
         dct_type += (int)slice->frame_wid - (int)slice->seq_hdr->horiz_size;
         wid      += (int)slice->frame_wid - (int)slice->seq_hdr->horiz_size;
         slice->ref_blkF += dct_type;
         slice->ref_blkB += dct_type;
      }
      else {
         dct_type = 0;
         wid *= 24 * (int)slice->cur_row;
      }
      wid  += (2 + (int)slice->cur_col) << 3;
      wid  -= (int)slice->luma_size;

      if (dct_type) {
         slice->cur_col = 0;
         slice->cur_row++;
      }
      else
         slice->cur_col++;
   }

   slice->pred_blk += wid;
   if (i < 5) {
      i++;
      goto START_LOOP;
   }
}
#endif

#undef HALF_PIXEL_X
#undef HALF_PIXEL_Y
#undef HALF_PIXEL_COLOUR_MASK
#undef HALF_PIXEL_COLOUR_SHIFT
#undef DCT_TOP_FIELD
#undef DCT_BOT_FIELD
#undef DCT_FRAME
#undef DCT_FRAME_TOP_LUMA
#undef DCT_FRAME_BOT_LUMA
#undef USE_SKIPPED_MACROBLOCK

