/***************************************************************************
                         oss_sound.h  -  description
                             -------------------
   Sound module implemented using OSS.
                             -------------------
    begin                : Wed May 11 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*#include <stdio.h>*/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include  <stdlib.h>
#include  <string.h>
#include  <unistd.h>
#include  <fcntl.h>
#ifdef OSS_NON_STANDARD
# include <soundcard.h>
#else
# include <sys/soundcard.h>
#endif

#ifdef _SPEED_TEST_
# include <stdio.h>
static int frame_cnt = 0;
#endif

#define  REQ_BLKSIZE_SHIFT 10
#define  REQ_BLKSIZE      ( 1 << REQ_BLKSIZE_SHIFT)
#define REQ_MAXBLKS        32
#define  REQ_BUFSIZE      (REQ_MAXBLKS * REQ_BLKSIZE)
#ifdef __LSB__
# define REQ_FORMAT        AFMT_S16_LE
#else
# define REQ_FORMAT        AFMT_S16_BE
#endif
#ifdef _NO_51_SURROUND_
# define MAX_CHANNELS      2
#else
# define MAX_CHANNELS      6
#endif
#ifndef OSS_DEFAULT_DEVICE
# define OSS_DEFAULT_DEVICE   "/dev/dsp"
#endif


const char *OSS_DEVICE = OSS_DEFAULT_DEVICE;
static int  audio_fd   = -1,
            oss_caps   =  0;
static u32  channels   =  0,
            rate       =  0,
            buf_size   =  0,
            buf_used   =  0;
static s16 *buf        = NULL;

/* Prepares the sound device for the incoming audio stream.
      INPUT: num_ch      - Number of audio channels.
             sample_rate - Sample rate (Hz)
      RETURN: Success - Number of audio channels.
              Failure - 0 */
u32 SOUND_init(u32 num_ch, u32 sample_rate)
{
#ifndef _SPEED_TEST_
   int arg;

   if (num_ch > 2)
      num_ch = MAX_CHANNELS;

   if (audio_fd >= 0) {
      if (!num_ch)
         goto SOUND_init_ERROR;
      else if ( (num_ch == channels)&&(sample_rate == rate) )
         return num_ch;
      close(audio_fd);
   }
   else if (!num_ch)
      return 0;

   audio_fd = open(OSS_DEVICE, O_WRONLY, 0);
   if (audio_fd < 0)
      return 0;   /* Opening device failed */

   if (!oss_caps) {
      if (ioctl(audio_fd, SNDCTL_DSP_GETCAPS, &oss_caps) < 0) {
         oss_caps = 0;
         goto SOUND_init_ERROR;  /* cannot determine device capabilities */
      }
   }

   if ((oss_caps & DSP_CAP_TRIGGER) == DSP_CAP_TRIGGER) {
      arg = ~PCM_ENABLE_OUTPUT;  /* disable playback */
      if (ioctl(audio_fd, SNDCTL_DSP_SETTRIGGER, &arg) < 0)
         goto SOUND_init_ERROR;  /* cannot disable playback */
   }
   else
      goto SOUND_init_ERROR;  /* cannot set trigger */

   arg = (REQ_MAXBLKS << 16) | REQ_BLKSIZE_SHIFT;
   if (ioctl(audio_fd, SNDCTL_DSP_SETFRAGMENT, &arg) < 0)
      goto SOUND_init_ERROR;  /* cannot set fragment size */

   arg = REQ_FORMAT;
   if ( (ioctl(audio_fd, SNDCTL_DSP_SETFMT, &arg) < 0)||(arg != REQ_FORMAT) )
      goto SOUND_init_ERROR;  /* cannot set format */


   arg = (int)num_ch;
   if ( (ioctl(audio_fd, SNDCTL_DSP_CHANNELS, &arg) < 0)||(arg != num_ch) ) {
      if (num_ch <= 2)
         goto SOUND_init_ERROR;     /* cannot set channels */
      else {
         /* cannot use 5.1 surround, so we're forced to live with stereo */
         if (arg != 2) {
            arg = 2;
            if ( (ioctl(audio_fd, SNDCTL_DSP_CHANNELS, &arg) < 0)||(arg != 2) )
               goto SOUND_init_ERROR;  /* cannot set channels */
         }
         num_ch = 2;
/*         printf("WARNING: Cannot use 5.1 surround, default to stereo\n");*/
      }
   }
   
   arg = (int)sample_rate;
   if ( (ioctl(audio_fd, SNDCTL_DSP_SPEED, &arg) < 0)||(arg != (int)sample_rate) )
      goto SOUND_init_ERROR;  /* cannot set sample rate */

   
   if (ioctl(audio_fd, SNDCTL_DSP_GETBLKSIZE, &arg) < 0)
      goto SOUND_init_ERROR;  /* cannot get block size */

   arg = PCM_ENABLE_OUTPUT;   /* enable playback */
   if (ioctl(audio_fd, SNDCTL_DSP_SETTRIGGER, &arg) < 0)
      goto SOUND_init_ERROR;  /* cannot enable playback */
#endif

/*   printf("SOUND_init(%u, %u)\n", num_ch, sample_rate);*/
   channels = num_ch;
   rate     = sample_rate;
   return num_ch;

#ifndef _SPEED_TEST_
SOUND_init_ERROR:
   close(audio_fd);
   audio_fd = -1;
   channels = rate = 0;
   return 0;
#endif
}

/* Returns the buffer that will hold the audio samples. Must
     call SOUND_init() function before calling this one.
      INPUT: samples - Number of samples the buffer will hold.
      RETURN:  Success - Pointer to the buffer.
               Failure - NULL */
s16 *SOUND_buf(u32 samples)
{
/*   printf("SOUND_buf(%u)", samples);*/
#ifndef _SPEED_TEST_
   if ( (audio_fd >= 0)&&(channels) ) {
#else
   if (channels) {
#endif
      samples *= sizeof(s16) * channels;
      if (samples > buf_size) {
         /* create a bigger buffer */
         if (buf)
            free(buf);
         buf      = malloc((size_t)samples);
         buf_size = (buf) ? samples : 0;
      }
      
      if (buf) {
         buf_used = samples;
         memset(buf, 0, (size_t)samples);
      }
      else
         buf_used = 0;
/*      printf("  buf_size=%u  buf_used=%u\n", buf_size, buf_used);*/
      return buf;
   }
   return NULL;
}

/* Called when the sound is no longer needed. */
void SOUND_done()
{
/*   printf("SOUND_done()\n");*/
#ifdef _SPEED_TEST_
   if (frame_cnt > 0) {
      rate *= channels * sizeof(s16);
      printf("AUDIO: FRAMES=%d  TARGET_FRAME_RATE=%.3f\n", 
             frame_cnt, ((double)rate / (double)buf_used));
      frame_cnt = 0;
      channels  = rate = 0;
   }
#endif
   if (audio_fd >= 0) {
      ioctl(audio_fd, SNDCTL_DSP_POST, 0);
      close(audio_fd);
      audio_fd = -1;
      channels = rate = 0;
   }
   if (buf) {
      free(buf);
      buf      = NULL;
      buf_size = 0;
   }
}

/* Called when the samples in the buffer should be played. */
void SOUND_play()
{
#ifndef _SPEED_TEST_
   write(audio_fd, buf, (size_t)buf_used);
#else
   frame_cnt++;
#endif
}

/* Returns the number of samples left in the buffer. */
u32 SOUND_samples_left()
{
#ifndef _USE_GETOSPACE_
   int bytes_left;

   if (ioctl(audio_fd, SNDCTL_DSP_GETODELAY, &bytes_left) < 0)
      return 0;   /* don't know how many bytes are left */
   return ((u32)bytes_left / (sizeof(s16) * channels));
#else
   audio_buf_info ret;

   if (ioctl(audio_fd, SNDCTL_DSP_GETOSPACE, &ret) < 0)
      return 0;   /* don't know how many bytes are left */
   ret.bytes = (ret.fragstotal*ret.fragsize) - ret.bytes;
   return ((u32)ret.bytes / (sizeof(s16) * channels));
#endif
}

/* This function is not needed when using OSS. */
void SOUND_stop(wait)
{
}

#undef BLKSIZE_SHIFT
#undef REQ_BLKSIZE
#undef REQ_MAXBLKS
#undef REQ_BUFSIZE
#undef REQ_FORMAT
#undef MAX_CHANNELS

